# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Process Git repository."""

from collections import deque

GRAY, BLACK = 0, 1


def topological(nodes):
    """Return nodes in a topological order."""
    order, enter, state = deque(), set(nodes), {}

    def dfs(node):
        """Visit nodes in depth-first order."""
        state[node] = GRAY
        for parent in nodes.get(node, ()):
            color = state.get(parent, None)
            if color == GRAY:
                raise ValueError('cycle')
            if color == BLACK:
                continue
            enter.discard(parent)
            dfs(parent)
        order.appendleft(node)
        state[node] = BLACK

    while enter:
        dfs(enter.pop())

    return order



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrap Docker API."""

import subprocess
from configparser import NoSectionError

import attr

from renku.core import errors
from renku.core.models.git import GitURL


def detect_registry_url(client, auto_login=True):
    """Return a URL of the Docker registry."""
    repo = client.repo
    config = repo.config_reader()

    # Find registry URL in .git/config
    remote_url = None
    try:
        registry_url = config.get_value('renku', 'registry', None)
    except NoSectionError:
        registry_url = None

    remote_branch = repo.head.reference.tracking_branch()

    if remote_branch is not None:
        remote_name = remote_branch.remote_name
        config_section = 'renku "{remote_name}"'.format(
            remote_name=remote_name
        )
        try:
            registry_url = config.get_value(
                config_section, 'registry', registry_url
            )
        except NoSectionError:
            pass
        remote_url = repo.remotes[remote_name].url

    if registry_url:
        # Look in [renku] and [renku "{remote_name}"] for registry_url key.
        url = GitURL.parse(registry_url)
    elif remote_url:
        # Use URL based on remote configuration.
        url = GitURL.parse(remote_url)

        # Replace gitlab. with registry. unless running on gitlab.com.
        hostname_parts = url.hostname.split('.')
        if len(hostname_parts) > 2 and hostname_parts[0] == 'gitlab':
            hostname_parts = hostname_parts[1:]
        hostname = '.'.join(['registry'] + hostname_parts)
        url = attr.evolve(url, hostname=hostname)
    else:
        raise errors.ConfigurationError(
            'Configure renku.repository_url or Git remote.'
        )

    if auto_login and url.username and url.password:
        try:
            subprocess.run([
                'docker',
                'login',
                url.hostname,
                '-u',
                url.username,
                '--password-stdin',
            ],
                           check=True,
                           input=url.password.encode('utf-8'))
        except subprocess.CalledProcessError:
            raise errors.AuthenticationError(
                'Check configuration of password or token in the registry URL'
            )

    return url



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Model objects representing projects."""

import datetime
import os

import attr

from renku.core.management.migrate import SUPPORTED_PROJECT_VERSION
from renku.core.models import jsonld
from renku.core.models.datastructures import Collection
from renku.core.models.provenance.agents import Person
from renku.core.utils.datetime8601 import parse_date

PROJECT_URL_PATH = 'projects'


@jsonld.s(
    type=[
        'schema:Project',
        'prov:Location',
    ],
    context={
        'schema': 'http://schema.org/',
        'prov': 'http://www.w3.org/ns/prov#'
    },
    slots=True,
)
class Project(object):
    """Represent a project."""

    name = jsonld.ib(default=None, context='schema:name')

    created = jsonld.ib(
        converter=parse_date,
        context='schema:dateCreated',
    )

    updated = jsonld.ib(
        converter=parse_date,
        context='schema:dateUpdated',
    )

    version = jsonld.ib(
        converter=str,
        default=str(SUPPORTED_PROJECT_VERSION),
        context='schema:schemaVersion',
    )

    client = attr.ib(default=None, kw_only=True)

    creator = jsonld.ib(
        default=None,
        kw_only=True,
        context={
            '@id': 'schema:creator',
        },
        type=Person
    )

    _id = jsonld.ib(context='@id', kw_only=True, default=None)

    @created.default
    @updated.default
    def _now(self):
        """Define default value for datetime fields."""
        return datetime.datetime.now(datetime.timezone.utc)

    def __attrs_post_init__(self):
        """Initialize computed attributes."""
        if not self.creator and self.client:
            if self.client.renku_metadata_path.exists():
                self.creator = Person.from_commit(
                    self.client.find_previous_commit(
                        self.client.renku_metadata_path, return_first=True
                    ),
                )
            else:
                # this assumes the project is being newly created
                self.creator = Person.from_git(self.client.repo)

        try:
            self._id = self.project_id
        except ValueError:
            """Fallback to old behaviour."""
            if not self._id and self.client and self.client.project:
                self._id = self.client.project._id

    @property
    def project_id(self):
        """Return the id for the project based on the repo origin remote."""
        import pathlib
        import urllib

        # Determine the hostname for the resource URIs.
        # If RENKU_DOMAIN is set, it overrides the host from remote.
        # Default is localhost.
        host = 'localhost'

        if not self.creator:
            raise ValueError('Project Creator not set')

        owner = self.creator.email.split('@')[0]
        name = self.name

        if self.client:
            remote = self.client.remote
            host = self.client.remote.get('host') or host
            owner = remote.get('owner') or owner
            name = remote.get('name') or name
        host = os.environ.get('RENKU_DOMAIN') or host
        if name:
            name = urllib.parse.quote(name, safe='')
        else:
            raise ValueError('Project name not set')

        project_url = urllib.parse.urljoin(
            'https://{host}'.format(host=host),
            pathlib.posixpath.join(PROJECT_URL_PATH, owner, name or 'NULL')
        )
        return project_url


class ProjectCollection(Collection):
    """Represent projects on the server.

    **Example**

    Create a project and check its name.

    # >>> project = client.projects.create(name='test-project')
    # >>> project.name
    # 'test-project'

    """

    class Meta:
        """Information about individual projects."""

        model = Project

    def create(self, name=None, **kwargs):
        """Create a new project.

        :param name: The name of the project.
        :returns: An instance of the newly create project.
        :rtype: renku.core.models.projects.Project
        """
        data = self._client.api.create_project({'name': name})
        return self.Meta.model(data, client=self._client, collection=self)

    def __getitem__(self, project_id):
        """Get an existing project by its id."""
        return self.Meta.model(
            self._client.api.get_project(project_id),
            client=self._client,
            collection=self
        )

    def __iter__(self):
        """Return all projects."""
        return (
            self.Meta.model(data, client=self._client, collection=self)
            for data in self._client.api.list_projects()
        )



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku save commands."""

from functools import reduce
from uuid import uuid4

import git

from renku.core import errors

from .client import pass_local_client


@pass_local_client
def save_and_push(client, message=None, remote=None, paths=None):
    """Save and push local changes."""
    client.setup_credential_helper()
    if not paths:
        paths = client.dirty_paths

    if paths:
        client.track_paths_in_storage(*paths)

    return repo_sync(client.repo, message, remote, paths)


def repo_sync(repo, message=None, remote=None, paths=None):
    """Commit and push paths."""
    origin = None
    saved_paths = []

    # get branch that's pushed
    if repo.active_branch.tracking_branch():
        pushed_branch = repo.active_branch.tracking_branch().name
    else:
        pushed_branch = repo.active_branch.name

    if remote:
        # get/setup supplied remote for pushing
        if repo.remotes:
            existing = next((r for r in repo.remotes if r.url == remote), None)
            if not existing:
                existing = next((r for r in repo.remotes if r.name == remote),
                                None)
            origin = next((r for r in repo.remotes if r.name == 'origin'),
                          None)
            if existing:
                origin = existing
            elif origin:
                pushed_branch = uuid4().hex
                origin = repo.create_remote(pushed_branch, remote)
        if not origin:
            origin = repo.create_remote('origin', remote)
    elif not repo.active_branch.tracking_branch():
        # No remote set on branch, push to available remote if only a single
        # one is available
        if len(repo.remotes) == 1:
            origin = repo.remotes[0]
        else:
            raise errors.ConfigurationError(
                'No remote has been set up for the current branch'
            )
    else:
        # get remote that's set up to track the local branch
        origin = repo.remotes[repo.active_branch.tracking_branch().remote_name]

    if paths:
        # commit uncommitted changes
        try:
            repo.git.add(*paths)
            saved_paths = [d.b_path for d in repo.index.diff('HEAD')]

            if not message:
                # Show saved files in message
                max_len = 100
                message = 'Saved changes to: '
                paths_with_lens = reduce(
                    lambda c, x: c + [(x, c[-1][1] + len(x))], saved_paths,
                    [(None, len(message))]
                )[1:]
                # limit first line to max_len characters
                message += ' '.join(
                    p if l < max_len else '\n\t' + p
                    for p, l in paths_with_lens
                )

            repo.index.commit(message)
        except git.exc.GitCommandError as e:
            raise errors.GitError('Cannot commit changes') from e

    try:
        # push local changes to remote branch
        origin.fetch()

        if origin.refs and repo.active_branch in origin.refs:
            origin.pull(repo.active_branch)

        origin.push(repo.active_branch)
    except git.exc.GitCommandError as e:
        if 'protected branches' not in e.stderr:
            raise errors.GitError('Cannot push changes') from e
        # push to new remote branch if original one is protected
        pushed_branch = uuid4().hex
        origin = repo.create_remote(pushed_branch, remote)
        origin.push(repo.active_branch)

    return saved_paths, pushed_branch



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test behavior of ``--output`` option."""

import os
from pathlib import Path

from renku.core.models.entities import Collection
from renku.core.models.provenance.activities import Activity


def read_all_workflow_files(client, glob='*.yaml'):
    """
    Return an array where its elements are content of CWL file
    found in the Renku project.
    """
    return [
        Activity.from_yaml(f, client=client)
        for f in client.workflow_path.glob(glob)
    ]


def test_run_succeeds_normally(cli, client, subdirectory):
    """Test when an output is detected"""
    foo = os.path.relpath(client.path / 'foo', os.getcwd())
    exit_code, cwl = cli('run', 'touch', foo)
    cwl = cwl.association.plan

    assert 0 == exit_code
    assert 0 == len(cwl.inputs)
    assert 1 == len(cwl.outputs)
    assert 'foo' == cwl.outputs[0].produces.path


def test_when_no_change_in_outputs_is_detected(cli, subdirectory):
    """Test when no output is detected"""
    cli('run', 'touch', 'foo')
    exit_code, cwl = cli('run', 'ls', 'foo')

    assert 1 == exit_code


def test_with_no_output_option(cli, client, subdirectory):
    """Test --no-output option with no output detection"""
    foo = os.path.relpath(client.path / 'foo', os.getcwd())
    cli('run', 'touch', foo)
    exit_code, cwl = cli('run', '--no-output', 'touch', foo)

    cwl = cwl.association.plan
    assert 0 == exit_code
    assert 1 == len(cwl.inputs)
    assert 'foo' == str(cwl.inputs[0].consumes.path)
    assert 0 == len(cwl.outputs)


def test_explicit_outputs_directory(cli, client, subdirectory):
    """Test detection of an output file with --output option."""
    foo = Path(os.path.relpath(client.path / 'foo', os.getcwd()))
    foo.mkdir()

    file_ = foo / 'file'

    exit_code, cwl = cli('run', '--output', str(foo), 'touch', str(file_))
    cwl = cwl.association.plan
    assert 0 == exit_code
    assert 0 == len(cwl.inputs)
    assert 1 == len(cwl.outputs)
    assert isinstance(cwl.outputs[0].produces, Collection)
    assert 'foo' == cwl.outputs[0].produces.path


def test_explicit_output_results(cli, client, subdirectory):
    """Test explicit output yield same results as normal run"""
    foo = os.path.relpath(client.path / 'foo', os.getcwd())
    cli('run', 'touch', foo)
    cli('run', '--output', foo, 'touch', foo)

    cwls = read_all_workflow_files(client)

    # There should be two command line tool.
    assert 2 == len(cwls)


def test_explicit_outputs_and_normal_outputs(cli, client, subdirectory):
    """Test explicit outputs and normal outputs can both exist"""
    foo = os.path.relpath(client.path / 'foo', os.getcwd())
    bar = os.path.relpath(client.path / 'bar', os.getcwd())
    cli('run', 'touch', foo)
    exit_code, cwl = cli('run', '--output', foo, 'touch', foo, bar)

    assert 0 == exit_code
    cwl = cwl.association.plan
    cwl.inputs.sort(key=lambda e: e.position)
    assert 2 == len(cwl.outputs)
    assert 'bar' == str(cwl.outputs[0].produces.path)
    assert 'foo' == str(cwl.outputs[1].produces.path)


def test_explicit_outputs_and_std_output_streams(cli, client, subdirectory):
    """Test that unchanged std output streams can be marked with explicit
    outputs"""
    exit_code, _ = cli('run', 'sh', '-c', 'echo foo > bar')
    assert 0 == exit_code

    exit_code, _ = cli('run', 'sh', '-c', 'echo foo > bar')
    assert 1 == exit_code

    exit_code, _ = cli('run', '--output', 'bar', 'sh', '-c', 'echo foo > bar')
    assert 0 == exit_code


def test_output_directory_with_output_option(cli, client, subdirectory):
    """Test output directories are not deleted with --output"""
    outdir = os.path.relpath(client.path / 'outdir', os.getcwd())
    a_script = ('sh', '-c', 'mkdir -p "$0"; touch "$0/$1"')
    cli('run', *a_script, outdir, 'foo')

    exit_code, _ = cli('run', '--output', outdir, *a_script, outdir, 'bar')

    assert 0 == exit_code
    assert (client.path / 'outdir' / 'foo').exists()
    assert (client.path / 'outdir' / 'bar').exists()


def test_output_directory_without_separate_outputs(cli, client):
    """Test output files not listed as separate outputs.

    See https://github.com/SwissDataScienceCenter/renku-python/issues/387
    """
    a_script = ('sh', '-c', 'mkdir -p "$0"; touch "$0/$1"')
    exit_code, cwl = cli('run', *a_script, 'outdir', 'foo')
    cwl = cwl.association.plan

    assert 0 == exit_code
    assert 1 == len(cwl.outputs)
    assert isinstance(cwl.outputs[0].produces, Collection)


def test_explicit_inputs_must_exist(cli):
    """Test explicit inputs exist before run"""
    exit_code, _ = cli('run', '--input', 'foo', 'touch', 'bar')

    assert 1 == exit_code


def test_explicit_inputs_are_inside_repo(cli):
    """Test explicit inputs are inside the Renku repo"""
    exit_code, _ = cli('run', '--input', '/tmp', 'touch', 'foo')

    assert 1 == exit_code


def test_explicit_inputs_and_outputs_are_listed(cli, client):
    """Test explicit inputs and outputs will be in generated CWL file"""
    foo = Path(os.path.relpath(client.path / 'foo', os.getcwd()))
    foo.mkdir()
    cli('run', 'touch', 'foo/file')
    cli('run', 'touch', 'bar', 'baz')

    exit_code, cwl = cli(
        'run', '--input', 'foo', '--input', 'bar', '--output', 'baz', 'echo'
    )
    assert 0 == exit_code

    cwl = cwl.association.plan
    assert 2 == len(cwl.inputs)
    cwl.inputs.sort(key=lambda e: e.consumes.path)

    assert cwl.inputs[0].position is None
    assert 'bar' == str(cwl.inputs[0].consumes.path)

    assert cwl.inputs[1].position is None
    assert 'foo' == str(cwl.inputs[1].consumes.path)
    assert isinstance(cwl.inputs[1].consumes, Collection)

    assert cwl.outputs[0].position is None
    assert not isinstance(cwl.outputs[0].produces, Collection)
    assert 'baz' == cwl.outputs[0].produces.path


def test_explicit_inputs_can_be_in_inputs(cli, client, subdirectory):
    """Test explicit inputs that are in inputs are treated as normal inputs"""
    foo = os.path.relpath(client.path / 'foo', os.getcwd())
    cli('run', 'touch', foo)

    exit_code, cwl = cli('run', '--input', foo, '--no-output', 'ls', foo)
    cwl = cwl.association.plan

    assert 0 == exit_code
    assert 1 == len(cwl.inputs)

    assert 'foo' == str(cwl.inputs[0].consumes.path)
    assert not isinstance(cwl.inputs[0].consumes, Collection)

    assert cwl.inputs[0].position is not None


def test_explicit_inputs_in_subdirectories(cli, client):
    """Test explicit inputs that are in sub-dirs are made accessible"""

    # Set up a script with hard dependency
    cli('run', '--no-output', 'mkdir', 'foo')
    cli('run', 'sh', '-c', 'echo "some changes" > foo/bar')
    cli('run', 'sh', '-c', 'echo "cat foo/bar" > script.sh')

    exit_code, _ = cli(
        'run', '--input', 'foo/bar', '--input', 'script.sh', 'sh', '-c',
        'sh script.sh > output'
    )
    assert 0 == exit_code

    # Status must be dirty if foo/bar changes
    cli('run', 'sh', '-c', 'echo "new changes" > foo/bar')
    exit_code, _ = cli('status')
    assert 1 == exit_code

    exit_code, _ = cli('update')
    assert 0 == exit_code
    assert (client.path / 'foo' / 'bar').exists()
    assert (client.path / 'script.sh').exists()
    assert (client.path / 'output').exists()



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent provenance entities."""

import os
import pathlib
import urllib
import weakref

import attr

from renku.core.models import jsonld as jsonld
from renku.core.models.projects import Project


def _str_or_none(data):
    """Return str representation or None."""
    return str(data) if data is not None else data


@attr.s(cmp=False)
class CommitMixin:
    """Represent a commit mixin."""

    commit = attr.ib(default=None, kw_only=True)
    client = attr.ib(default=None, kw_only=True)
    path = jsonld.ib(
        context='prov:atLocation',
        default=None,
        kw_only=True,
        converter=_str_or_none
    )

    _id = jsonld.ib(default=None, context='@id', kw_only=True)
    _label = jsonld.ib(context='rdfs:label', kw_only=True)
    _project = jsonld.ib(
        context='schema:isPartOf', type=Project, kw_only=True, default=None
    )

    @property
    def submodules(self):
        """Proxy to client submodules."""
        if self.client:
            return self.client.submodules

    def default_id(self):
        """Configure calculated ID."""
        if self.commit:
            hexsha = self.commit.hexsha
        else:
            hexsha = 'UNCOMMITTED'

        # Determine the hostname for the resource URIs.
        # If RENKU_DOMAIN is set, it overrides the host from remote.
        # Default is localhost.
        host = 'localhost'
        if self.client:
            host = self.client.remote.get('host') or host
        host = os.environ.get('RENKU_DOMAIN') or host

        # always set the id by the identifier
        return urllib.parse.urljoin(
            'https://{host}'.format(host=host),
            pathlib.posixpath.join(
                '/blob/{hexsha}/{path}'.format(hexsha=hexsha, path=self.path)
            )
        )

    @_label.default
    def default_label(self):
        """Generate a default label."""
        if self.commit:
            hexsha = self.commit.hexsha
        else:
            hexsha = 'UNCOMMITTED'
        if self.path:
            path = self.path
            if self.client and os.path.isabs(path):
                path = pathlib.Path(path).relative_to(self.client.path)
            return '{path}@{hexsha}'.format(hexsha=hexsha, path=path)
        return '{hexsha}'.format(hexsha=hexsha, self=self)

    def __attrs_post_init__(self):
        """Post-init hook."""
        if self.path and self.client:
            path = pathlib.Path(self.path)
            if path.is_absolute():
                self.path = str(path.relative_to(self.client.path))

        # always force "project" to be the current project
        if self.client:
            self._project = self.client.project

        if not self._id:
            self._id = self.default_id()


@jsonld.s(
    type=[
        'prov:Entity',
        'wfprov:Artifact',
    ],
    context={
        'schema': 'http://schema.org/',
        'prov': 'http://www.w3.org/ns/prov#',
        'wfprov': 'http://purl.org/wf4ever/wfprov#',
        'rdfs': 'http://www.w3.org/2000/01/rdf-schema#'
    },
    cmp=False,
)
class Entity(CommitMixin):
    """Represent a data value or item."""

    _parent = attr.ib(
        default=None,
        kw_only=True,
        converter=lambda value: weakref.ref(value)
        if value is not None else None,
    )

    @classmethod
    def from_revision(
        cls,
        client,
        path,
        revision='HEAD',
        parent=None,
        find_previous=True,
        **kwargs
    ):
        """Return dependency from given path and revision."""
        if find_previous:
            revision = client.find_previous_commit(path, revision=revision)

        client, commit, path = client.resolve_in_submodules(
            revision,
            path,
        )

        path_ = client.path / path
        if path != '.' and path_.is_dir():
            entity = Collection(
                client=client,
                commit=commit,
                path=path,
                members=[],
                parent=parent,
            )

            files_in_commit = commit.stats.files

            # update members with commits
            for member in path_.iterdir():
                if member.name == '.gitkeep':
                    continue

                member_path = str(member.relative_to(client.path))
                find_previous = True

                if member_path in files_in_commit:
                    # we already know the newest commit, no need to look it up
                    find_previous = False

                try:
                    assert all(member_path != m.path for m in entity.members)

                    entity.members.append(
                        cls.from_revision(
                            client,
                            member_path,
                            commit,
                            parent=entity,
                            find_previous=find_previous,
                            **kwargs
                        )
                    )
                except KeyError:
                    pass

        else:
            entity = cls(
                client=client,
                commit=commit,
                path=str(path),
                parent=parent,
                **kwargs
            )

        return entity

    @property
    def parent(self):  # pragma: no cover
        """Return the parent object."""
        return self._parent() if self._parent is not None else None

    @property
    def entities(self):
        """Yield itself."""
        if (
            self.client and not self.commit and self._label and
            '@UNCOMMITTED' not in self._label
        ):
            self.commit = self.client.repo.commit(self._label.rsplit('@')[1])

        yield self

    def set_client(self, client):
        """Sets the clients on this entity."""
        self.client = client


@jsonld.s(
    type=[
        'prov:Collection',
    ],
    context={
        'prov': 'http://www.w3.org/ns/prov#',
    },
    cmp=False,
)
class Collection(Entity):
    """Represent a directory with files."""

    members = jsonld.container.list(
        type=Entity, context='prov:hadMember', kw_only=True
    )

    def default_members(self):
        """Generate default members as entities from current path."""
        if not self.client:
            return []
        dir_path = self.client.path / self.path

        if not dir_path.exists():
            # likely a directory deleted in a previous commit
            return []

        assert dir_path.is_dir()

        members = []
        for path in dir_path.iterdir():
            if path.name == '.gitkeep':
                continue  # ignore empty directories in Git repository
            cls = Collection if path.is_dir() else Entity
            members.append(
                cls(
                    commit=self.commit,
                    client=self.client,
                    path=str(path.relative_to(self.client.path)),
                    parent=self,
                )
            )
        return members

    @property
    def entities(self):
        """Recursively return all files."""
        for member in self.members:
            if not member.client and self.client:
                member.client = self.client
            yield from member.entities

        if (
            self.client and not self.commit and self._label and
            '@UNCOMMITTED' not in self._label
        ):
            self.commit = self.client.repo.commit(self._label.rsplit('@')[1])

        yield self

    def set_client(self, client):
        """Sets the clients on this entity."""
        super().set_client(client)

        for m in self.members:
            m.set_client(client)

    def __attrs_post_init__(self):
        """Init members."""
        super().__attrs_post_init__()

        if self.members is None:
            self.members = self.default_members()

        for member in self.members:
            member._parent = weakref.ref(self)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service views."""
from flask import current_app

from renku.service.serializers.rpc import JsonRPCResponse


def result_response(serializer, data):
    """Construct flask response."""
    return current_app.response_class(
        response=serializer.dumps({'result': data}),
        mimetype='application/json'
    )


def error_response(code, reason):
    """Construct error response."""
    return current_app.response_class(
        response=JsonRPCResponse().dumps({
            'error': {
                'code': code,
                'reason': reason
            }
        }),
        mimetype='application/json'
    )



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent a Git commit."""

import os
import urllib
import uuid
import weakref
from collections import OrderedDict
from pathlib import Path, posixpath

import attr
from git import NULL_TREE

from renku.core.models import jsonld
from renku.core.models.cwl.annotation import Annotation
from renku.core.models.entities import Collection, CommitMixin, Entity
from renku.core.models.refs import LinkReference
from renku.core.models.workflow.run import Run

from .agents import Person, renku_agent
from .qualified import Association, Generation, Usage


def _nodes(output, parent=None):
    """Yield nodes from entities."""
    # NOTE refactor so all outputs behave the same
    entity = getattr(output, 'entity', output)

    if isinstance(entity, Collection):
        for member in entity.members:
            if parent is not None:
                member = attr.evolve(member, parent=parent)

            if entity.client:
                _set_entity_client_commit(member, entity.client, None)
            if isinstance(output, Generation):
                child = Generation(
                    activity=output.activity,
                    entity=member,
                    role=entity.role if hasattr(entity, 'role') else None
                )
            elif isinstance(output, Usage):
                child = Usage(
                    activity=output.activity,
                    entity=member,
                    role=entity.role if hasattr(entity, 'role') else None
                )
            else:
                child = member
            yield from _nodes(child)

    yield output


def _set_entity_client_commit(entity, client, commit):
    """Set the client and commit of an entity."""
    if client and not entity.client:
        entity.client = client

    if not entity.commit:
        revision = 'UNCOMMITTED'
        if entity._label:
            revision = entity._label.rsplit('@')[1]
        if revision == 'UNCOMMITTED':
            commit = commit
        elif client:
            commit = client.repo.commit(revision)
        entity.commit = commit


@jsonld.s(
    type='prov:Activity',
    context={
        'prov': 'http://www.w3.org/ns/prov#',
        'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
        'schema': 'http://schema.org/'
    },
    cmp=False,
)
class Activity(CommitMixin):
    """Represent an activity in the repository."""

    _id = jsonld.ib(default=None, context='@id', kw_only=True)
    _message = jsonld.ib(context='rdfs:comment', kw_only=True)
    _was_informed_by = jsonld.ib(
        context='prov:wasInformedBy',
        kw_only=True,
    )

    part_of = attr.ib(default=None, kw_only=True)

    _collections = attr.ib(
        default=attr.Factory(OrderedDict), init=False, kw_only=True
    )
    generated = jsonld.container.list(
        Generation, context={
            '@reverse': 'prov:activity',
        }, kw_only=True
    )

    invalidated = jsonld.container.list(
        Entity, context={
            '@reverse': 'prov:wasInvalidatedBy',
        }, kw_only=True
    )

    influenced = jsonld.ib(
        context='prov:influenced',
        kw_only=True,
    )

    started_at_time = jsonld.ib(
        context={
            '@id': 'prov:startedAtTime',
            '@type': 'http://www.w3.org/2001/XMLSchema#dateTime',
        },
        kw_only=True,
    )

    ended_at_time = jsonld.ib(
        context={
            '@id': 'prov:endedAtTime',
            '@type': 'http://www.w3.org/2001/XMLSchema#dateTime',
        },
        kw_only=True,
    )

    agent = jsonld.ib(
        context='prov:wasAssociatedWith',
        kw_only=True,
        default=renku_agent,
        type='renku.core.models.provenance.agents.SoftwareAgent'
    )
    person_agent = jsonld.ib(
        context='prov:wasAssociatedWith',
        kw_only=True,
        type='renku.core.models.provenance.agents.Person'
    )

    def default_generated(self):
        """Create default ``generated``."""
        generated = []

        for path in self.get_output_paths():
            entity = self._get_activity_entity(path)

            generated.append(
                Generation(activity=self, entity=entity, role=None)
            )
        return generated

    def get_output_paths(self):
        """Gets all output paths generated by this run."""
        index = set()

        commit = self.commit

        if not self.commit:
            if not self.client:
                return index
            commit = self.client.repo.head.commit

        for file_ in commit.diff(commit.parents or NULL_TREE):
            # ignore deleted files (note they appear as ADDED)
            # in this backwards diff
            if file_.change_type == 'A':
                continue
            path_ = Path(file_.a_path)

            is_dataset = self.client.DATASETS in str(path_)
            not_refs = LinkReference.REFS not in str(path_)
            does_not_exists = not path_.exists()

            if all([is_dataset, not_refs, does_not_exists]):
                uid = uuid.UUID(path_.parent.name)
                path_ = (
                    Path(self.client.renku_home) / self.client.DATASETS /
                    str(uid) / self.client.METADATA
                )

            index.add(str(path_))

        return index

    def _get_activity_entity(self, path, deleted=False):
        """Gets the entity associated with this Activity and path."""
        client, commit, path = self.client.resolve_in_submodules(
            self.commit,
            path,
        )
        output_path = client.path / path
        parents = list(output_path.relative_to(client.path).parents)

        collection = None
        members = []
        for parent in reversed(parents[:-1]):
            if str(parent) in self._collections:
                collection = self._collections[str(parent)]
            else:
                collection = Collection(
                    client=client,
                    commit=commit,
                    path=str(parent),
                    members=[],
                    parent=collection,
                )
                members.append(collection)
                self._collections[str(parent)] = collection

            members = collection.members

        entity_cls = Entity
        if (self.client.path / path).is_dir():
            entity_cls = Collection

        # TODO: use a factory method to generate the entity
        if str(path).startswith(
            os.path.join(client.renku_home, client.DATASETS)
        ) and not deleted:
            entity = client.load_dataset_from_path(path, commit=commit)
        else:
            entity = entity_cls(
                commit=commit,
                client=client,
                path=str(path),
                parent=collection,
            )

        if collection:
            collection.members.append(entity)

        return entity

    def default_invalidated(self):
        """Entities invalidated by this Action."""
        results = []
        for path in self.removed_paths:
            entity = self._get_activity_entity(path, deleted=True)

            results.append(entity)
        return results

    @influenced.default
    def default_influenced(self):
        """Calculate default values."""
        return list(self._collections.values())

    @property
    def parents(self):
        """Return parent commits."""
        if self.commit:
            return list(self.commit.parents)

    @property
    def removed_paths(self):
        """Return all paths removed in the commit."""
        index = set()
        if not self.commit:
            return index

        for file_ in self.commit.diff(self.commit.parents or NULL_TREE):
            # only process deleted files (note they appear as ADDED)
            # in this backwards diff
            if file_.change_type != 'A':
                continue
            path_ = Path(file_.a_path)

            index.add(str(path_))

        return index

    @property
    def paths(self):
        """Return all paths in the commit."""
        index = set()

        for file_ in self.commit.diff(self.commit.parents or NULL_TREE):
            # ignore deleted files (note they appear as ADDED)
            # in this backwards diff
            if file_.change_type == 'A':
                continue
            path_ = Path(file_.a_path)

            is_dataset = self.client.DATASETS in str(path_)
            not_refs = LinkReference.REFS not in str(path_)
            does_not_exists = not (
                path_.exists() or
                (path_.is_symlink() and os.path.lexists(path_))
            )

            if all([is_dataset, not_refs, does_not_exists]):
                uid = uuid.UUID(path_.parent.name)
                path_ = (
                    Path(self.client.renku_home) / self.client.DATASETS /
                    str(uid) / self.client.METADATA
                )

            index.add(str(path_))

        return index

    @classmethod
    def generate_id(cls, commitsha):
        """Calculate action ID."""
        host = 'localhost'
        if hasattr(cls, 'client'):
            host = cls.client.remote.get('host') or host
        host = os.environ.get('RENKU_DOMAIN') or 'localhost'

        # always set the id by the identifier
        return urllib.parse.urljoin(
            'https://{host}'.format(host=host),
            posixpath.join(
                '/activities', 'commit/{commit}'.format(commit=commitsha)
            )
        )

    def default_id(self):
        """Configure calculated ID."""
        if self.commit:
            return self.generate_id(self.commit.hexsha)
        return self.generate_id('UNCOMMITED')

    @_message.default
    def default_message(self):
        """Generate a default message."""
        if self.commit:
            return self.commit.message

    @_was_informed_by.default
    def default_was_informed_by(self):
        """List parent actions."""
        if self.commit:
            return [{
                '@id': self.generate_id(parent),
            } for parent in self.commit.parents]

    @started_at_time.default
    def default_started_at_time(self):
        """Configure calculated properties."""
        if self.commit:
            return self.commit.authored_datetime.isoformat()

    @ended_at_time.default
    def default_ended_at_time(self):
        """Configure calculated properties."""
        if self.commit:
            return self.commit.committed_datetime.isoformat()

    @person_agent.default
    def default_person_agent(self):
        """Set person agent to be the author of the commit."""
        if self.commit:
            return Person.from_commit(self.commit)
        return None

    @property
    def nodes(self):
        """Return topologically sorted nodes."""
        collections = OrderedDict()

        def _parents(node):
            if node.parent:
                yield from _parents(node.parent)
                yield node.parent

        for output in self.generated:
            for parent in _parents(output.entity):
                collections[parent.path] = parent

            yield from _nodes(output)

        for removed in self.invalidated:
            for parent in _parents(removed):
                collections[parent.path] = parent

            yield from _nodes(removed)

        yield from reversed(collections.values())

    def __attrs_post_init__(self):
        """Sets ``generated`` and ``invalidated`` default values if needed."""
        super().__attrs_post_init__()
        if not self._id:
            self._id = self.default_id()
        if not self.generated:
            self.generated = self.default_generated()

        for g in self.generated:
            _set_entity_client_commit(g.entity, self.client, self.commit)

        if not self.invalidated:
            self.invalidated = self.default_invalidated()

        if self.generated:
            for g in self.generated:
                g._activity = weakref.ref(self)


@jsonld.s(
    type='wfprov:ProcessRun',
    context={
        'wfprov': 'http://purl.org/wf4ever/wfprov#',
        'oa': 'http://www.w3.org/ns/oa#',
    },
    cmp=False,
)
class ProcessRun(Activity):
    """A process run is a particular execution of a Process description."""

    __association_cls__ = Run

    generated = jsonld.container.list(
        Generation,
        context={
            '@reverse': 'prov:activity',
        },
        kw_only=True,
        default=None
    )

    association = jsonld.ib(
        context='prov:qualifiedAssociation',
        default=None,
        kw_only=True,
        type=Association
    )

    annotations = jsonld.container.list(
        context={
            '@reverse': 'oa:hasTarget',
        }, kw_only=True, type=Annotation
    )

    qualified_usage = jsonld.container.list(
        Usage, context='prov:qualifiedUsage', kw_only=True, default=None
    )

    def __attrs_post_init__(self):
        """Calculate properties."""
        super().__attrs_post_init__()

        if not self.commit and self.client:
            self.commit = self.client.find_previous_commit(self.path)

        if not self.annotations:
            self.annotations = self.plugin_annotations()

        if self.association:
            self.association.plan._activity = weakref.ref(self)
            plan = self.association.plan
            if not plan.commit:
                if self.client:
                    plan.client = self.client
                if self.commit:
                    plan.commit = self.commit

                if plan.inputs:
                    for i in plan.inputs:
                        _set_entity_client_commit(
                            i.consumes, self.client, self.commit
                        )
                if plan.outputs:
                    for o in plan.outputs:
                        _set_entity_client_commit(
                            o.produces, self.client, self.commit
                        )

        if self.qualified_usage and self.client and self.commit:
            usages = []
            revision = '{0}'.format(self.commit)
            for usage in self.qualified_usage:
                if not usage.commit and '@UNCOMMITTED' in usage._label:
                    usages.append(
                        Usage.from_revision(
                            client=self.client,
                            path=usage.path,
                            role=usage.role,
                            revision=revision,
                            id=usage._id,
                        )
                    )
                else:
                    if not usage.client:
                        usage.entity.set_client(self.client)
                    if not usage.commit:
                        revision = usage._label.rsplit('@')[1]
                        usage.entity.commit = self.client.repo.commit(revision)

                    usages.append(usage)
            self.qualified_usage = usages

    def default_generated(self):
        """Create default ``generated``."""
        generated = []

        if not self.association or not self.association.plan:
            return generated

        for output in self.association.plan.outputs:
            entity = Entity.from_revision(
                self.client,
                output.produces.path,
                revision=self.commit,
                parent=output.produces.parent
            )

            generation = Generation(
                activity=self, role=output.sanitized_id, entity=entity
            )
            generated.append(generation)
        return generated

    def add_annotations(self, annotations):
        """Adds annotations from an external tool."""
        self.annotations.extend(annotations)

    def plugin_annotations(self):
        """Adds ``Annotation``s from plugins to a ``ProcessRun``."""
        from renku.core.plugins.pluginmanager import get_plugin_manager
        pm = get_plugin_manager()

        results = pm.hook.process_run_annotations(run=self)
        return [a for r in results for a in r]

    @classmethod
    def from_run(
        cls,
        run,
        client,
        path,
        commit=None,
        is_subprocess=False,
        update_commits=False
    ):
        """Convert a ``Run`` to a ``ProcessRun``."""
        from .agents import SoftwareAgent

        if not commit:
            commit = client.repo.head.commit

        usages = []

        id_ = ProcessRun.generate_id(commit)

        if is_subprocess:
            id_ = '{}/steps/step_{}'.format(id_, run.process_order)

        for input_ in run.inputs:
            usage_id = id_ + '/inputs/' + input_.sanitized_id
            revision = commit
            input_path = input_.consumes.path
            entity = input_.consumes
            if update_commits:
                revision = client.find_previous_commit(
                    input_path, revision=commit.hexsha
                )
                entity = Entity.from_revision(client, input_path, revision)

            dependency = Usage(
                entity=entity, role=input_.sanitized_id, id=usage_id
            )

            usages.append(dependency)

        agent = SoftwareAgent.from_commit(commit)
        association = Association(
            agent=agent, id=id_ + '/association', plan=run
        )

        process_run = cls(
            id=id_,
            qualified_usage=usages,
            association=association,
            client=client,
            commit=commit,
            path=path
        )

        generated = []

        for output in run.outputs:
            entity = Entity.from_revision(
                client,
                output.produces.path,
                revision=commit,
                parent=output.produces.parent
            )

            generation = Generation(
                activity=process_run, role=output.sanitized_id, entity=entity
            )
            generated.append(generation)

        process_run.generated = generated

        process_run.plugin_annotations()
        return process_run

    @property
    def parents(self):
        """Return parent commits."""
        return [
            member.commit for usage in self.qualified_usage
            for member in usage.entity.entities
        ] + super().parents

    @property
    def nodes(self):
        """Return topologically sorted nodes."""
        # Outputs go first
        yield from super().nodes

        # Activity itself
        yield self.association.plan


@jsonld.s(
    type='wfprov:WorkflowRun',
    context={
        'wfprov': 'http://purl.org/wf4ever/wfprov#',
    },
    cmp=False,
)
class WorkflowRun(ProcessRun):
    """A workflow run typically contains several subprocesses."""

    __association_cls__ = Run

    _processes = jsonld.container.list(
        ProcessRun,
        context={
            '@reverse': 'wfprov:wasPartOfWorkflowRun',
        },
        kw_only=True,
        default=attr.Factory(list)
    )

    @property
    def subprocesses(self):
        """Subprocesses of this ``WorkflowRun``."""
        return {p.association.plan.process_order: p for p in self._processes}

    @classmethod
    def from_run(cls, run, client, path, commit=None, update_commits=False):
        """Convert a ``Run`` to a ``WorkflowRun``."""
        from .agents import SoftwareAgent

        if not commit:
            commit = client.repo.head.commit

        processes = []
        generated = []

        for s in run.subprocesses:
            proc_run = ProcessRun.from_run(
                s, client, path, commit, True, update_commits
            )
            processes.append(proc_run)
            generated.extend(proc_run.generated)

        usages = []

        id_ = cls.generate_id(commit)

        for input in run.inputs:
            usage_id = id_ + '/inputs/' + input.sanitized_id
            dependency = Usage.from_revision(
                client=client,
                path=input.consumes.path,
                role=input.sanitized_id,
                revision=commit,
                id=usage_id,
            )
            usages.append(dependency)

        agent = SoftwareAgent.from_commit(commit)
        association = Association(
            agent=agent, id=id_ + '/association', plan=run
        )

        all_generated = []

        # fix generations in folders
        for generation in generated:
            all_generated.append(generation)
            entity = generation.entity

            if not isinstance(entity, Collection) or not entity.commit:
                continue

            for e in entity.entities:
                if e.commit is not entity.commit:
                    continue

                all_generated.append(
                    Generation(
                        activity=generation.activity, entity=e, role=None
                    )
                )

        wf_run = WorkflowRun(
            id=id_,
            processes=processes,
            generated=all_generated,
            qualified_usage=usages,
            association=association,
            client=client,
            commit=commit,
            path=path
        )
        return wf_run

    @property
    def nodes(self):
        """Yield all graph nodes."""
        for subprocess in reversed(self._processes):
            if subprocess.path is None:
                # skip nodes connecting directory to file
                continue

            for n in subprocess.nodes:
                # if self.client and not n.commit and isinstance(n, Entity):
                #     _set_entity_client_commit(n, self.client, self.commit)
                # n._activity = weakref.ref(subprocess)
                yield n
            yield subprocess.association.plan

    def __attrs_post_init__(self):
        """Attrs post initializations."""
        if not self._id:
            self._id = self.default_id()

        if self.client:
            for s in self._processes:
                s.client = self.client
                s.commit = self.commit
                s.__attrs_post_init__()
                s.part_of = self

        super().__attrs_post_init__()



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Track provenance of data created by executing programs.

Capture command line execution
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Tracking execution of your command line script is done by simply adding the
``renku run`` command before the actual command. This will enable detection of:

* arguments (flags),
* string and integer options,
* input files or directories if linked to existing paths in the repository,
* output files or directories if modified or created while running the command.

.. note:: If there were uncommitted changes in the repository, then the
   ``renku run`` command fails. See :program:`git status` for details.

.. warning:: Input and output paths can only be detected if they are passed as
   arguments to ``renku run``.

.. warning:: Circular dependencies are not supported for ``renku run``. See
   :ref:`circular-dependencies` for more details.

.. warning:: When using output redirection in ``renku run`` on Windows (with
   `` > file`` or `` 2> file``), all Renku errors and messages are redirected
   as well and ``renku run`` produces no output on the terminal. On Linux,
   this is detected by renku and only the output of the command to be run is
   actually redirected. Renku specific messages such as errors get printed to
   the terminal as usual and don't get redirected.

Detecting input paths
~~~~~~~~~~~~~~~~~~~~~

Any path passed as an argument to ``renku run``, which was not changed during
the execution, is identified as an input path. The identification only works if
the path associated with the argument matches an existing file or directory
in the repository.

The detection might not work as expected
if:

* a file is **modified** during the execution. In this case it will be stored
  as an **output**;
* a path is not passed as an argument to ``renku run``.

.. topic:: Specifying auxiliary inputs (``--input``)

   You can specify extra inputs to your program explicitly by using the
   ``--input`` option. This is useful for specifying hidden dependencies
   that don't appear on the command line. These input file must exist before
   execution of ``renku run`` command. This option is not a replacement for
   the arguments that are passed on the command line. Files or directories
   specified with this option will not be passed as input arguments to the
   script.

Detecting output paths
~~~~~~~~~~~~~~~~~~~~~~

Any path **modified** or **created** during the execution will be added as an
output.

Because the output path detection is based on the Git repository state after
the execution of ``renku run`` command, it is good to have a basic
understanding of the underlying principles and limitations of tracking
files in Git.

Git tracks not only the paths in a repository, but also the content stored in
those paths. Therefore:

* a recreated file with the same content is not considered an output file,
  but instead is kept as an input;
* file moves are detected based on their content and can cause problems;
* directories cannot be empty.

.. note:: When in doubt whether the outputs will be detected, remove all
  outputs using ``git rm <path>`` followed by ``git commit`` before running
  the ``renku run`` command.

.. topic:: Command does not produce any files (``--no-output``)

   If the program does not produce any outputs, the execution ends with an
   error:

   .. code-block:: text

      Error: There are not any detected outputs in the repository.

   You can specify the ``--no-output`` option to force tracking of such
   an execution.

.. topic:: Specifying outputs explicitly (``--output``)

   You can specify expected outputs of your program explicitly by using the
   ``--output`` option. These output must exist after the execution of the
   ``renku run`` command. However, they do not need to be modified by
   the command.

.. cli-run-std

Detecting standard streams
~~~~~~~~~~~~~~~~~~~~~~~~~~

Often the program expect inputs as a standard input stream. This is detected
and recorded in the tool specification when invoked by ``renku run cat < A``.

Similarly, both redirects to standard output and standard error output can be
done when invoking a command:

.. code-block:: console

   $ renku run grep "test" B > C 2> D

.. warning:: Detecting inputs and outputs from pipes ``|`` is not supported.

Specifying inputs and outputs programmatically
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Sometimes the list of inputs and outputs are not known before execution of the
program. For example, a program might accept a date range as input and access
all files within that range during its execution.

To address this issue, the program can dump a list of input and output files
that it is accessing in ``inputs.txt`` and ``outputs.txt``. Each line in these
files is expected to be the path to an input or output file within the
project's directory. When the program is finished, Renku will look for
existence of these two files and adds their content to the list of explicit
inputs and outputs. Renku will then delete these two files.

By default, Renku looks for these two files in ``.renku/tmp`` directory. One
can change this default location by setting ``RENKU_FILELIST_PATH``
environment variable. When set, it points to the directory within the
project's directory where ``inputs.txt`` and ``outputs.txt`` reside.

Exit codes
~~~~~~~~~~

All Unix commands return a number between 0 and 255 which is called
"exit code". In case other numbers are returned, they are treaded module 256
(-10 is equivalent to 246, 257 is equivalent to 1). The exit-code 0 represents
a *success* and non-zero exit-code indicates a *failure*.

Therefore the command specified after ``renku run`` is expected to return
exit-code 0. If the command returns different exit code, you can specify them
with ``--success-code=<INT>`` parameter.

.. code-block:: console

   $ renku run --success-code=1 --no-output fail

.. _circular-dependencies:

Circular Dependencies
~~~~~~~~~~~~~~~~~~~~~

Circular dependencies are not supported in ``renku run``. This means you cannot
use the same file or directory as both an input and an output in the same step,
for instance reading from a file as input and then appending to it is not
allowed. Since renku records all steps of an analysis workflow in a dependency
graph and it allows you to update outputs when an input changes, this would
lead to problems with circular dependencies. An update command would change the
input again, leading to renku seeing it as a changed input, which would run
update again, and so on, without ever stopping.

Due to this, the renku depedency graph has to be *acyclic*. So instead of
appending to an input file or writing an output file to the same directory
that was used as an input directory, create new files or write to other
directories, respectively.
"""

import os
import sys
from subprocess import call

import click

from renku.core import errors
from renku.core.commands.client import pass_local_client
from renku.core.commands.options import option_isolation
from renku.core.management.git import _mapped_std_streams
from renku.core.models.cwl.command_line_tool import CommandLineToolFactory


@click.command(context_settings=dict(ignore_unknown_options=True, ))
@click.option(
    'inputs',
    '--input',
    multiple=True,
    help='Force a path to be considered as an input.',
)
@click.option(
    'outputs',
    '--output',
    multiple=True,
    help='Force a path to be considered an output.',
)
@click.option(
    '--no-output',
    is_flag=True,
    default=False,
    help='Allow command without output files.',
)
@click.option(
    '--success-code',
    'success_codes',
    type=int,
    multiple=True,
    callback=lambda _, __, values: [int(value) % 256 for value in values],
    help='Allowed command exit-code.',
)
@option_isolation
@click.argument('command_line', nargs=-1, type=click.UNPROCESSED)
@pass_local_client(
    clean=True,
    requires_migration=True,
    up_to_date=True,
    commit=True,
    ignore_std_streams=True,
)
def run(
    client, inputs, outputs, no_output, success_codes, isolation, command_line
):
    """Tracking work on a specific problem."""
    mapped_std = _mapped_std_streams(client.candidate_paths)
    system_stdout = None
    system_stderr = None

    # /dev/tty is a virtual device that points to the terminal
    # of the currently executed process
    try:
        with open('/dev/tty', 'w'):
            tty_exists = True
    except OSError:
        tty_exists = False

    try:
        stdout_redirected = 'stdout' in mapped_std
        stderr_redirected = 'stderr' in mapped_std

        if tty_exists:
            # if renku was called with redirected stdout/stderr, undo the
            # redirection here so error messages can be printed normally
            if stdout_redirected:
                system_stdout = open('/dev/tty', 'w')
                old_stdout = sys.stdout
                sys.stdout = system_stdout

            if stderr_redirected:
                system_stderr = open('/dev/tty', 'w')
                old_stderr = sys.stderr
                sys.stderr = system_stderr

        working_dir = client.repo.working_dir
        factory = CommandLineToolFactory(
            command_line=command_line,
            explicit_inputs=inputs,
            explicit_outputs=outputs,
            directory=os.getcwd(),
            working_dir=working_dir,
            successCodes=success_codes,
            **{
                name: os.path.relpath(path, working_dir)
                for name, path in mapped_std.items()
            }
        )
        with client.with_workflow_storage() as wf:
            with factory.watch(client, no_output=no_output) as tool:
                # Don't compute paths if storage is disabled.
                if client.check_external_storage():
                    # Make sure all inputs are pulled from a storage.
                    paths_ = (
                        path for _, path in
                        tool.iter_input_files(client.workflow_path)
                    )
                    client.pull_paths_from_storage(*paths_)

                if tty_exists:
                    # apply original output redirection
                    if stdout_redirected:
                        sys.stdout = old_stdout
                    if stderr_redirected:
                        sys.stderr = old_stderr

                return_code = call(
                    factory.command_line,
                    cwd=os.getcwd(),
                    **{key: getattr(sys, key)
                       for key in mapped_std.keys()},
                )

                sys.stdout.flush()
                sys.stderr.flush()

                if tty_exists:
                    # change back to /dev/tty redirection
                    if stdout_redirected:
                        sys.stdout = system_stdout
                    if stderr_redirected:
                        sys.stderr = system_stderr

                if return_code not in (success_codes or {0}):
                    raise errors.InvalidSuccessCode(
                        return_code, success_codes=success_codes
                    )

                wf.add_step(run=tool)

        if factory.messages:
            click.echo(factory.messages)

        if factory.warnings:
            click.echo(factory.warnings)

    finally:
        if system_stdout:
            sys.stdout = old_stdout
            system_stdout.close()
        if system_stderr:
            sys.stderr = old_stderr
            system_stderr.close()



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CLI tests."""

from __future__ import absolute_import, print_function

import contextlib
import json
import os
import subprocess
import sys
from pathlib import Path

import git
import pytest
from click.testing import CliRunner
from cwlgen import parse_cwl
from tests.cli.test_init import INPUT, METADATA, TEMPLATE_ID

from renku import __version__
from renku.cli import cli
from renku.core.management.storage import StorageApiMixin
from renku.core.utils.contexts import chdir


def test_version(runner):
    """Test cli version."""
    result = runner.invoke(cli, ['--version'])
    assert __version__ in result.output.split('\n')


@pytest.mark.parametrize('arg', (('help', ), ('-h', ), ('--help', )))
def test_help(arg, runner):
    """Test cli help."""
    result = runner.invoke(cli, [arg])
    assert 0 == result.exit_code
    assert 'Show this message and exit.' in result.output


@pytest.mark.parametrize('cwd', ('data', 'notebooks', 'subdir'))
def test_run_from_non_root(runner, client, cwd):
    path = client.path / cwd
    path.mkdir(parents=True, exist_ok=True)
    with chdir(path):
        result = runner.invoke(cli, ['dataset'])
        assert 0 == result.exit_code
        assert 'Run CLI commands only from project\'s root' in result.output

        result = runner.invoke(cli, ['help'])
        assert 0 == result.exit_code
        assert 'Run CLI commands only from project' not in result.output

    result = runner.invoke(cli, ['dataset'])
    assert 0 == result.exit_code
    assert 'Run CLI commands only from project\'s root' not in result.output


def test_config_path(runner):
    """Test config path."""
    result = runner.invoke(cli, ['--global-config-path'])
    output = result.output.split('\n')[0]
    assert 'renku.ini' in output


def test_show_context(runner, project):
    """Test context generation."""
    result = runner.invoke(cli, ['show', 'context', '--list'])
    contexts = [name for name in result.output.split('\n') if name]

    assert 0 == result.exit_code
    assert 1 < len(contexts)

    result = runner.invoke(cli, ['show', 'context'] + contexts)
    assert 0 == result.exit_code

    data = json.loads(result.output)
    assert len(contexts) == len(data)


_CMD_EXIT_2 = ['bash', '-c', 'exit 2']


@pytest.mark.parametrize(
    'cmd, exit_code', (
        (_CMD_EXIT_2, 1),
        (['--success-code', '1', '--no-output'] + _CMD_EXIT_2, 1),
        (['--success-code', '2', '--no-output'] + _CMD_EXIT_2, 0),
        (['--success-code', '0', '--no-output', 'echo', 'hola'], 0),
    )
)
def test_exit_code(cmd, exit_code, runner, project):
    """Test exit-code of run command."""
    result = runner.invoke(cli, ['run'] + cmd)
    assert exit_code == result.exit_code


def test_git_pre_commit_hook(runner, project, capsys):
    """Test detection of output edits."""
    result = runner.invoke(cli, ['githooks', 'install'])
    assert 0 == result.exit_code
    assert 'Hook already exists.' in result.output

    repo = git.Repo(project)
    cwd = Path(project)
    output = cwd / 'output.txt'

    result = runner.invoke(cli, ['run', 'touch', output.name])
    assert 0 == result.exit_code
    with output.open('w') as f:
        f.write('hello')

    repo.git.add('--all')
    with pytest.raises(git.HookExecutionError) as error:
        repo.index.commit('hello')
        assert output.name in error.stdout

    result = runner.invoke(cli, ['githooks', 'uninstall'])
    assert 0 == result.exit_code

    repo.index.commit('hello')


def test_workflow(runner, project):
    """Test workflow command."""
    result = runner.invoke(cli, ['run', 'touch', 'data.csv'])
    assert 0 == result.exit_code

    with open('counted.txt', 'w') as stdout:
        with contextlib.redirect_stdout(stdout):
            try:
                cli.main(
                    args=('run', 'wc', 'data.csv'),
                    prog_name=runner.get_default_prog_name(cli),
                )
            except SystemExit as e:
                assert e.code in {None, 0}

    result = runner.invoke(
        cli,
        ['workflow', 'create', 'counted.txt', '-o', 'workflow.cwl'],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code
    workflow = parse_cwl('workflow.cwl')
    assert 2 == len(workflow.steps)

    # Compare default log and log for a specific file.
    result_default = runner.invoke(cli, ['log'])
    result_arg = runner.invoke(cli, ['log', 'counted.txt'])

    assert 0 == result_default.exit_code
    assert 0 == result_arg.exit_code
    assert result_default.output == result_arg.output


def test_streams(runner, project, capsys, no_lfs_warning):
    """Test redirection of std streams."""
    repo = git.Repo('.')

    with open('source.txt', 'w') as source:
        source.write('first,second,third')

    repo.git.add('--all')
    repo.index.commit('Added source.txt')

    with capsys.disabled():
        with open('source.txt', 'rb') as stdin:
            with open('result.txt', 'wb') as stdout:
                try:
                    old_stdin, old_stdout = sys.stdin, sys.stdout
                    sys.stdin, sys.stdout = stdin, stdout
                    try:
                        cli.main(
                            args=('run', 'cut', '-d,', '-f', '2', '-s'),
                            prog_name=runner.get_default_prog_name(cli),
                        )
                    except SystemExit as e:
                        assert e.code in {None, 0}
                finally:
                    sys.stdin, sys.stdout = old_stdin, old_stdout

    with open('result.txt', 'r') as f:
        assert f.read().strip() == 'second'

    result = runner.invoke(cli, ['workflow', 'create', 'result.txt'])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ['status'])
    assert 0 == result.exit_code

    # Check that source.txt is not shown in outputs.
    result = runner.invoke(cli, ['show', 'outputs', 'source.txt'])
    assert 1 == result.exit_code

    result = runner.invoke(cli, ['show', 'outputs'])
    assert 0 == result.exit_code
    assert {
        'result.txt',
    } == set(result.output.strip().split('\n'))

    # Check that source.txt is shown in inputs.
    result = runner.invoke(cli, ['show', 'inputs'])
    assert 0 == result.exit_code
    assert {
        'source.txt',
    } == set(result.output.strip().split('\n'))

    with open('source.txt', 'w') as source:
        source.write('first,second,third,fourth')

    repo.git.add('--all')
    repo.index.commit('Changed source.txt')

    result = runner.invoke(cli, ['status'])
    assert 1 == result.exit_code
    assert 'source.txt' in result.output


def test_streams_cleanup(runner, project, run):
    """Test cleanup of standard streams."""
    source = Path(project) / 'source.txt'
    stdout = Path(project) / 'result.txt'

    with source.open('w') as fp:
        fp.write('first,second,third')

    # File outside the Git index should be deleted.

    with source.open('r') as fp:
        assert fp.read() == 'first,second,third'

    assert not stdout.exists()

    result = runner.invoke(cli, ['status'])

    # Dirty repository check.
    assert 1 == result.exit_code

    # File from the Git index should be restored.
    repo = git.Repo(project)
    with stdout.open('w') as fp:
        fp.write('1')

    repo.index.add(['result.txt'])

    with stdout.open('r') as fp:
        assert fp.read() == '1'


def test_streams_and_args_names(runner, project, capsys, no_lfs_warning):
    """Test streams and conflicting argument names."""
    with capsys.disabled():
        with open('lalala', 'wb') as stdout:
            try:
                old_stdout = sys.stdout
                sys.stdout = stdout
                try:
                    cli.main(args=('run', 'echo', 'lalala'), )
                except SystemExit as e:
                    assert e.code in {None, 0}
            finally:
                sys.stdout = old_stdout

    with open('lalala', 'r') as f:
        assert f.read().strip() == 'lalala'

    result = runner.invoke(cli, ['status'], catch_exceptions=False)
    assert 0 == result.exit_code


def test_show_inputs(tmpdir_factory, project, runner, run):
    """Test show inputs with submodules."""
    second_project = Path(str(tmpdir_factory.mktemp('second_project')))

    parameters = []
    for key in set(METADATA.keys()):
        parameters.append('--parameter')
        parameters.append(f'{key}="{METADATA[key]}"')
    assert 0 == run(
        args=(
            'init', str(second_project), '--template-id', TEMPLATE_ID,
            *parameters
        )
    )

    woop = second_project / 'woop'
    with woop.open('w') as fp:
        fp.write('woop')

    second_repo = git.Repo(str(second_project))
    second_repo.git.add('--all')
    second_repo.index.commit('Added woop file')

    assert 0 == run(args=('dataset', 'create', 'foo'))
    assert 0 == run(args=('dataset', 'add', 'foo', str(woop)))

    imported_woop = Path(project) / 'data' / 'foo' / woop.name
    assert imported_woop.exists()

    woop_wc = Path(project) / 'woop.wc'
    assert 0 == run(args=('run', 'wc'), stdin=imported_woop, stdout=woop_wc)

    result = runner.invoke(cli, ['show', 'inputs'], catch_exceptions=False)
    assert {str(imported_woop.resolve().relative_to(Path(project).resolve()))
            } == set(result.output.strip().split('\n'))


def test_configuration_of_no_external_storage(isolated_runner, monkeypatch):
    """Test the LFS requirement for renku run with disabled storage."""
    runner = isolated_runner

    os.mkdir('test-project')
    os.chdir('test-project')

    result = runner.invoke(
        cli,
        ['--no-external-storage', 'init', '.', '--template-id', TEMPLATE_ID],
        INPUT
    )
    assert 0 == result.exit_code
    # Pretend that git-lfs is not installed.
    with monkeypatch.context() as monkey:
        monkey.setattr(StorageApiMixin, 'storage_installed', False)
        # Missing --no-external-storage flag.
        result = runner.invoke(cli, ['run', 'touch', 'output'])
        assert 'External storage is not configured' in result.output
        assert 1 == result.exit_code

        # Since repo is not using external storage.
        result = runner.invoke(
            cli, ['--no-external-storage', 'run', 'touch', 'output']
        )
        assert 0 == result.exit_code

    subprocess.call(['git', 'clean', '-df'])
    result = runner.invoke(
        cli, ['--no-external-storage', 'run', 'touch', 'output']
    )
    # Needs to result in error since output file
    # is now considered an input file (check run.py doc).
    assert 1 == result.exit_code


def test_configuration_of_external_storage(isolated_runner, monkeypatch):
    """Test the LFS requirement for renku run."""
    runner = isolated_runner

    result = runner.invoke(
        cli, ['--external-storage', 'init', '.', '--template-id', TEMPLATE_ID],
        INPUT
    )
    assert 0 == result.exit_code
    # Pretend that git-lfs is not installed.
    with monkeypatch.context() as monkey:
        monkey.setattr(StorageApiMixin, 'storage_installed', False)
        # Repo is using external storage but it's not installed.
        result = runner.invoke(cli, ['run', 'touch', 'output'])
        assert 1 == result.exit_code
        assert 'External storage is not configured' in result.output

        result = runner.invoke(cli, ['-S', 'run', 'touch', 'output'])
        assert 1 == result.exit_code
        assert 'External storage is not installed' in result.output

    # Clean repo and check external storage.
    subprocess.call(['git', 'clean', '-df'])
    result = runner.invoke(cli, ['run', 'touch', 'output2'])
    assert 0 == result.exit_code


def test_early_check_of_external_storage(
    isolated_runner, monkeypatch, directory_tree
):
    """Test LFS is checked early."""
    result = isolated_runner.invoke(
        cli,
        ['--no-external-storage', 'init', '.', '--template-id', TEMPLATE_ID]
    )
    assert 0 == result.exit_code

    result = isolated_runner.invoke(cli, ['dataset', 'create', 'my-dataset'])
    assert 0 == result.exit_code

    # Pretend that git-lfs is not installed.
    with monkeypatch.context() as monkey:
        monkey.setattr(StorageApiMixin, 'storage_installed', False)

        failing_command = [
            'dataset', 'add', '-s', 'src', 'my-dataset', directory_tree.strpath
        ]
        result = isolated_runner.invoke(cli, failing_command)
        assert 1 == result.exit_code
        assert 'External storage is not configured' in result.output

        result = isolated_runner.invoke(
            cli, ['--no-external-storage'] + failing_command
        )
        assert 2 == result.exit_code
        assert 'Cannot use "--source" with URLs' in result.output


def test_file_tracking(isolated_runner):
    """Test .gitattribute handling on renku run."""
    runner = isolated_runner

    os.mkdir('test-project')
    os.chdir('test-project')
    result = runner.invoke(
        cli, ['init', '.', '--template-id', TEMPLATE_ID], INPUT
    )
    assert 0 == result.exit_code
    result = runner.invoke(cli, ['config', 'lfs_threshold', '0b'])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ['run', 'touch', 'tracked'])
    assert 0 == result.exit_code
    assert 'tracked' in Path('.gitattributes').read_text()

    result = runner.invoke(cli, ['-S', 'run', 'touch', 'untracked'])
    assert 0 == result.exit_code
    assert 'untracked' not in Path('.gitattributes').read_text()


@pytest.mark.xfail
def test_status_with_submodules(isolated_runner, monkeypatch):
    """Test status calculation with submodules."""
    runner = isolated_runner

    os.mkdir('foo')
    os.mkdir('bar')

    with open('woop', 'w') as f:
        f.write('woop')

    os.chdir('foo')
    result = runner.invoke(
        cli, ['init', '.', '--template', TEMPLATE_ID, '--no-external-storage'],
        INPUT,
        catch_exceptions=False
    )
    assert 0 == result.exit_code

    os.chdir('../bar')
    result = runner.invoke(
        cli, ['init', '.', '--template', TEMPLATE_ID, '--no-external-storage'],
        INPUT,
        catch_exceptions=False
    )
    assert 0 == result.exit_code

    os.chdir('../foo')
    with monkeypatch.context() as monkey:
        monkey.setattr(StorageApiMixin, 'storage_installed', False)

        result = runner.invoke(
            cli, ['dataset', 'add', 'f', '../woop'], catch_exceptions=False
        )

        assert 1 == result.exit_code
        subprocess.call(['git', 'clean', '-dff'])

    result = runner.invoke(
        cli, ['-S', 'dataset', 'add', 'f', '../woop'], catch_exceptions=False
    )
    assert 0 == result.exit_code

    os.chdir('../bar')
    result = runner.invoke(
        cli, ['-S', 'dataset', 'add', 'b', '../foo/data/f/woop'],
        catch_exceptions=False
    )
    assert 0 == result.exit_code

    # Produce a derived data from the imported data.
    with open('woop.wc', 'w') as stdout:
        with contextlib.redirect_stdout(stdout):
            try:
                cli.main(
                    args=('-S', 'run', 'wc', 'data/b/woop'),
                    prog_name=runner.get_default_prog_name(cli),
                )
            except SystemExit as e:
                assert e.code in {None, 0}

    result = runner.invoke(cli, ['status'], catch_exceptions=False)
    assert 0 == result.exit_code

    # Modify the source data.
    os.chdir('../foo')
    with open('data/f/woop', 'w') as f:
        f.write('woop2')

    subprocess.call(['git', 'commit', '-am', 'commiting changes to woop'])

    os.chdir('../bar')
    subprocess.call(['git', 'submodule', 'update', '--rebase', '--remote'])
    subprocess.call(['git', 'commit', '-am', 'update submodule'])

    result = runner.invoke(cli, ['status'], catch_exceptions=False)
    assert 0 != result.exit_code

    # Test relative log output
    cmd = ['--path', '../foo', 'log']
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    assert '../foo/data/f/woop' in result.output
    assert 0 == result.exit_code


def test_status_consistency(client, project):
    """Test if the renku status output is consistent when running the
    command from directories other than the repository root."""
    runner = CliRunner(mix_stderr=False)

    os.mkdir('somedirectory')
    with open('somedirectory/woop', 'w') as fp:
        fp.write('woop')

    client.repo.index.add(['somedirectory/woop'])
    client.repo.index.commit('add woop')

    result = runner.invoke(
        cli, ['run', 'cp', 'somedirectory/woop', 'somedirectory/meeh']
    )
    assert 0 == result.exit_code

    with open('somedirectory/woop', 'w') as fp:
        fp.write('weep')

    client.repo.index.add(['somedirectory/woop'])
    client.repo.index.commit('fix woop')

    base_result = runner.invoke(cli, ['status'])
    os.chdir('somedirectory')
    comp_result = runner.invoke(cli, ['status'])
    assert base_result.stdout.replace(
        'somedirectory/', ''
    ) == comp_result.output


def test_unchanged_output(runner, project):
    """Test detection of unchanged output."""
    cmd = ['run', 'touch', '1']
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    assert 0 == result.exit_code

    cmd = ['run', 'touch', '1']
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    assert 1 == result.exit_code


def test_unchanged_stdout(runner, project, capsys, no_lfs_warning):
    """Test detection of unchanged stdout."""
    with capsys.disabled():
        with open('output.txt', 'wb') as stdout:
            try:
                old_stdout = sys.stdout
                sys.stdout = stdout
                try:
                    cli.main(args=('run', 'echo', '1'), )
                except SystemExit as e:
                    assert e.code in {None, 0}
            finally:
                sys.stdout = old_stdout

    with capsys.disabled():
        with open('output.txt', 'wb') as stdout:
            try:
                old_stdout = sys.stdout
                sys.stdout = stdout
                try:
                    cli.main(args=('run', 'echo', '1'), )
                except SystemExit as e:
                    # The stdout has not been modified!
                    assert e.code in {None, 1}
            finally:
                sys.stdout = old_stdout


def test_modified_output(runner, project, run):
    """Test detection of changed file as output."""
    cwd = Path(project)
    source = cwd / 'source.txt'
    data = cwd / 'data' / 'results'
    data.mkdir(parents=True)
    output = data / 'result.txt'

    repo = git.Repo(project)
    cmd = ['run', 'cp', '-r', str(source), str(output)]

    def update_source(data):
        """Update source.txt."""
        with source.open('w') as fp:
            fp.write(data)

        repo.git.add('--all')
        repo.index.commit('Updated source.txt')

    update_source('1')

    # The output file does not exist.
    assert not output.exists()

    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code

    # The output file is copied from the source.
    with output.open('r') as f:
        assert f.read().strip() == '1'

    update_source('2')

    # The input file has been updated and output is recreated.
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code

    with output.open('r') as f:
        assert f.read().strip() == '2'

    update_source('3')

    # The input has been modifed and we check that the previous
    # run command correctly recognized output.txt.
    assert 0 == run()

    with output.open('r') as f:
        assert f.read().strip() == '3'


def test_siblings(runner, project):
    """Test detection of siblings."""
    siblings = {'brother', 'sister'}

    cmd = ['run', 'touch'] + list(siblings)
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code

    for sibling in siblings:
        cmd = ['show', 'siblings', sibling]
        result = runner.invoke(cli, cmd)
        assert 0 == result.exit_code

        output = {
            name.strip()
            for name in result.output.split('\n') if name.strip()
        }
        assert output == siblings, 'Checked {0}'.format(sibling)


def test_orphan(runner, project):
    """Test detection of an orphan."""
    cwd = Path(project)
    orphan = cwd / 'orphan.txt'

    cmd = ['run', 'touch', orphan.name]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code

    cmd = ['show', 'siblings', 'orphan.txt']
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code
    assert 'orphan.txt\n' == result.output


def test_only_child(runner, project):
    """Test detection of an only child."""
    cmd = ['run', 'touch', 'only_child']
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code

    cmd = ['show', 'siblings', 'only_child']
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code
    assert 'only_child\n' == result.output


def test_outputs(runner, project):
    """Test detection of outputs."""
    siblings = {'brother', 'sister'}

    cmd = ['run', 'touch'] + list(siblings)
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code

    result = runner.invoke(cli, ['show', 'outputs'])
    assert 0 == result.exit_code
    assert siblings == set(result.output.strip().split('\n'))


def test_moved_file(runner, project):
    """Test that moved files are displayed correctly."""
    repo = git.Repo(project)
    cwd = Path(project)
    input_ = cwd / 'input.txt'
    with input_.open('w') as f:
        f.write('first')

    repo.git.add('--all')
    repo.index.commit('Created input.txt')

    result = runner.invoke(cli, ['log'])
    assert 0 == result.exit_code
    assert input_.name in result.output

    repo.git.mv(input_.name, 'renamed.txt')
    repo.index.commit('Renamed input')

    result = runner.invoke(cli, ['log'])
    assert 0 == result.exit_code
    assert input_.name not in result.output
    assert 'renamed.txt' in result.output


def test_deleted_input(runner, project, capsys):
    """Test deleted input."""
    repo = git.Repo(project)
    cwd = Path(project)
    input_ = cwd / 'input.txt'
    with input_.open('w') as f:
        f.write('first')

    repo.git.add('--all')
    repo.index.commit('Created input.txt')

    cmd = ['run', 'mv', input_.name, 'input.mv']
    result = runner.invoke(cli, cmd, catch_exceptions=False)
    assert 0 == result.exit_code
    assert not input_.exists()
    assert Path('input.mv').exists()


def test_input_directory(runner, project, run, no_lfs_warning):
    """Test detection of input directory."""
    repo = git.Repo(project)
    cwd = Path(project)
    output = cwd / 'output.txt'
    inputs = cwd / 'inputs'
    inputs.mkdir(parents=True)

    gitkeep = inputs / '.gitkeep'
    gitkeep.touch()
    repo.git.add('--all')
    repo.index.commit('Empty inputs directory')

    assert 0 == run(args=('run', 'ls', str(inputs)), stdout=output)
    with output.open('r') as fp:
        assert '' == fp.read().strip()

    (inputs / 'first').touch()

    repo.git.add('--all')
    repo.index.commit('Created inputs')

    assert 0 == run(args=('update', output.name))

    with output.open('r') as fp:
        assert 'first\n' == fp.read()

    (inputs / 'second').touch()
    repo.git.add('--all')
    repo.index.commit('Added second input')

    assert 0 == run(args=('update', output.name))
    with output.open('r') as fp:
        assert 'first\nsecond\n' == fp.read()

    result = runner.invoke(cli, ['show', 'inputs'])
    assert 0 == result.exit_code
    assert set(
        str(p.relative_to(cwd))
        for p in inputs.rglob('*') if p.name != '.gitkeep'
    ) == set(result.output.strip().split('\n'))


@pytest.mark.parametrize(
    'global_only,config_path_attr',
    ((False, 'local_config_path'), (True, 'global_config_path'))
)
def test_config_manager_creation(
    client, global_config_dir, global_only, config_path_attr
):
    """Check creation of configuration file."""
    path = getattr(client, config_path_attr)
    assert path.endswith('renku.ini')
    config = client.load_config(local_only=False, global_only=False)
    client.store_config(config, global_only=global_only)
    assert Path(path).exists()


@pytest.mark.parametrize('global_only', (False, True))
def test_config_manager_set_value(client, global_config_dir, global_only):
    """Check writing to configuration."""
    local_only = not global_only

    client.set_value(
        'zenodo', 'access_token', 'my-secret', global_only=global_only
    )

    config = client.load_config(local_only=local_only, global_only=global_only)
    assert config.get('zenodo', 'access_token') == 'my-secret'

    client.remove_value('zenodo', 'access_token', global_only=global_only)
    config = client.load_config(local_only=local_only, global_only=global_only)
    assert 'zenodo' not in config.sections()


def test_config_get_value(client, global_config_dir):
    """Check reading from configuration."""
    # Value set locally is not visible globally
    client.set_value('local', 'key', 'local-value')
    value = client.get_value('local', 'key')
    assert 'local-value' == value
    value = client.get_value('local', 'key', global_only=True)
    assert value is None

    # Value set globally is stored globally
    client.set_value('global', 'key', 'global-value', global_only=True)
    value = client.get_value('global', 'key', local_only=True)
    assert value is None
    value = client.get_value('global', 'key', global_only=True)
    assert 'global-value' == value
    value = client.get_value('global', 'key')
    assert 'global-value' == value

    # Reading non-existing values returns None
    value = client.get_value('non-existing', 'key')
    assert value is None


def test_lfs_size_limit(isolated_runner):
    """Test inclusion of files in lfs by size."""
    runner = isolated_runner

    os.mkdir('test-project')
    os.chdir('test-project')
    result = runner.invoke(
        cli, ['init', '.', '--template-id', TEMPLATE_ID], INPUT
    )
    assert 0 == result.exit_code

    large = Path('large')
    with large.open('w') as f:
        f.write('x' * 1024**2)

    result = runner.invoke(
        cli, ['dataset', 'add', '--create', 'my-dataset',
              str(large)],
        catch_exceptions=False
    )
    assert 0 == result.exit_code
    assert 'large' in Path('.gitattributes').read_text()

    small = Path('small')
    with small.open('w') as f:
        f.write('small file')

    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(small)],
        catch_exceptions=False
    )
    assert 0 == result.exit_code
    assert 'small' not in Path('.gitattributes').read_text()


@pytest.mark.parametrize(
    'ignore,path,tracked',
    (('file1', 'file1', False), ('!file1', 'file1', True),
     ('dir1', 'dir1/file1', False), ('dir1/file1', 'dir1/file1', False),
     ('dir1\n!dir1/file1', 'dir1/file1', True), ('*.test', 'file.test', False),
     ('*.test', 'file.txt', True), ('*.test', 'dir2/file.test', False),
     ('dir2\n!*.test', 'dir2/file.test', True))
)
def test_lfs_ignore(isolated_runner, ignore, path, tracked):
    """Test inclusion of files in lfs by size."""
    runner = isolated_runner

    os.mkdir('test-project')
    os.chdir('test-project')
    result = runner.invoke(
        cli, ['init', '.', '--template-id', TEMPLATE_ID], INPUT
    )
    assert 0 == result.exit_code
    result = runner.invoke(cli, ['config', 'lfs_threshold', '0b'])
    assert 0 == result.exit_code

    with Path('.renkulfsignore').open('w') as f:
        f.write(ignore)
    subprocess.call(['git', 'commit', '-am', 'Add lfs ignore'])

    # force creation of .gitattributes by adding tracked dummy file
    with Path('dummy').open('w') as f:
        f.write('dummy')

    result = runner.invoke(
        cli, ['dataset', 'add', '--create', 'my-dataset', 'dummy'],
        catch_exceptions=False
    )

    # add dataset file
    filepath = Path(path)

    if str(filepath.parent) != '.':
        filepath.parent.mkdir(parents=True, exist_ok=True)

    with filepath.open('w') as f:
        f.write('x' * 1024**2)

    result = runner.invoke(
        cli, ['dataset', 'add', 'my-dataset',
              str(filepath)],
        catch_exceptions=False
    )
    assert 0 == result.exit_code

    # check if file is tracked or not
    if tracked:
        assert str(filepath) in Path('.gitattributes').read_text()
    else:
        assert str(filepath) not in Path('.gitattributes').read_text()



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test compliance with ``.gitignore`` file."""
from renku.cli import cli


def test_dataset_add(tmpdir, runner, client, subdirectory):
    """Test importing data into a dataset."""
    # create a dataset
    result = runner.invoke(cli, ['dataset', 'create', 'testing'])
    assert 0 == result.exit_code
    assert 'OK' in result.output

    # Using an extension from gitignore.default defined as *.spec
    ignored_file = tmpdir.join('my.spec')
    ignored_file.write('My Specification')

    # The file should be ignored and command fail
    result = runner.invoke(
        cli,
        ['dataset', 'add', 'testing', ignored_file.strpath],
        catch_exceptions=False,
    )

    assert 1 == result.exit_code

    client.repo.git.clean('-dff')

    # Use the --force ;)
    result = runner.invoke(
        cli, ['dataset', 'add', 'testing', '--force', ignored_file.strpath]
    )
    assert 0 == result.exit_code



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""JSON-LD dataset migrations."""
import itertools
import json
import os
import uuid
from pathlib import Path

import pyld

from renku.core.models.jsonld import read_yaml, write_yaml


def migrate(client):
    """Migration function."""
    _migrate_project_metadata(client)
    _migrate_datasets_metadata(client)


def _migrate_project_metadata(client):
    """Apply all initial JSON-LD migrations to project."""
    jsonld_translate = {
        'http://schema.org/name': 'http://xmlns.com/foaf/0.1/name',
        'http://schema.org/Project': 'http://xmlns.com/foaf/0.1/Project'
    }

    _apply_on_the_fly_jsonld_migrations(
        path=client.renku_metadata_path,
        jsonld_context=_INITIAL_JSONLD_PROJECT_CONTEXT,
        fields=_PROJECT_FIELDS,
        jsonld_translate=jsonld_translate
    )


def _migrate_datasets_metadata(client):
    """Apply all initial JSON-LD migrations to datasets."""
    jsonld_migrations = {
        'dctypes:Dataset': [_migrate_dataset_schema, _migrate_absolute_paths],
        'schema:Dataset': [
            _migrate_absolute_paths, _migrate_doi_identifier,
            _migrate_same_as_structure, _migrate_dataset_file_id
        ],
    }

    old_metadata_paths = _dataset_pre_0_3(client)
    new_metadata_paths = client.renku_datasets_path.rglob(client.METADATA)

    for path in itertools.chain(old_metadata_paths, new_metadata_paths):
        _apply_on_the_fly_jsonld_migrations(
            path=path,
            jsonld_context=_INITIAL_JSONLD_DATASET_CONTEXT,
            fields=_DATASET_FIELDS,
            client=client,
            jsonld_migrations=jsonld_migrations
        )


def _apply_on_the_fly_jsonld_migrations(
    path,
    jsonld_context,
    fields,
    client=None,
    jsonld_migrations=None,
    jsonld_translate=None
):
    data = read_yaml(path)

    if jsonld_translate:
        # perform the translation

        data = pyld.jsonld.expand(data)
        data_str = json.dumps(data)
        for k, v in jsonld_translate.items():
            data_str = data_str.replace(v, k)
        data = json.loads(data_str)
        data = pyld.jsonld.compact(data, jsonld_context)

    data.setdefault('@context', jsonld_context)

    if jsonld_migrations:
        schema_type = data.get('@type')
        migrations = []

        if isinstance(schema_type, list):
            for schema in schema_type:
                migrations += jsonld_migrations.get(schema, [])
        elif isinstance(schema_type, str):
            migrations += jsonld_migrations.get(schema_type, [])

        for migration in set(migrations):
            data = migration(data, client)

    if data['@context'] != jsonld_context:
        # merge new context into old context to prevent properties
        # getting lost in jsonld expansion
        if isinstance(data['@context'], str):
            data['@context'] = {'@base': data['@context']}
        data['@context'].update(jsonld_context)
        try:
            compacted = pyld.jsonld.compact(data, jsonld_context)
        except Exception:
            compacted = data
    else:
        compacted = data

    data = {}

    for k, v in compacted.items():
        if k in fields:
            no_value_context = isinstance(v, dict) and '@context' not in v
            has_nested_context = (
                k in compacted['@context'] and
                '@context' in compacted['@context'][k]
            )
            if no_value_context and has_nested_context:
                # Propagate down context
                v['@context'] = compacted['@context'][k]['@context']

            data[k] = v

    write_yaml(path, data)


def _dataset_pre_0_3(client):
    """Return paths of dataset metadata for pre 0.3.4."""
    project_is_pre_0_3 = int(client.project.version) < 2
    if project_is_pre_0_3:
        return (client.path / 'data').rglob(client.METADATA)
    return []


def _migrate_dataset_schema(data, client):
    """Migrate from old dataset formats."""
    if 'authors' not in data:
        return

    data['@context']['creator'] = data['@context'].pop(
        'authors', {'@container': 'list'}
    )

    data['creator'] = data.pop('authors', {})

    files = data.get('files', [])

    if isinstance(files, dict):
        files = files.values()
    for file_ in files:
        file_['creator'] = file_.pop('authors', {})

    return data


def _migrate_absolute_paths(data, client):
    """Migrate dataset paths to use relative path."""
    raw_path = data.get('path', '.')
    path = Path(raw_path)

    if path.is_absolute():
        try:
            data['path'] = str(path.relative_to(os.getcwd()))
        except ValueError:
            elements = raw_path.split('/')
            index = elements.index('.renku')
            data['path'] = str(Path('/'.join(elements[index:])))

    files = data.get('files', [])

    if isinstance(files, dict):
        files = list(files.values())

    for file_ in files:
        path = Path(file_.get('path'), '.')
        if path.is_absolute():
            file_['path'] = str(path.relative_to((os.getcwd())))
    data['files'] = files
    return data


def _migrate_doi_identifier(data, client):
    """If the dataset _id is doi, make it a UUID."""
    from renku.core.utils.doi import is_doi
    from renku.core.utils.uuid import is_uuid

    _id = data.get('_id', '')
    identifier = data.get('identifier', '')

    if not is_uuid(_id):
        if not is_uuid(identifier):
            data['identifier'] = str(uuid.uuid4())
        if is_doi(data.get('_id', '')):
            data['same_as'] = {'@type': ['schema:URL'], 'url': data['_id']}
            if data.get('@context'):
                data['@context'].setdefault(
                    'same_as', {
                        '@id': 'schema:sameAs',
                        '@type': 'schema:URL',
                        '@context': {
                            '@version': '1.1',
                            'url': 'schema:url',
                            'schema': 'http://schema.org/'
                        }
                    }
                )
        data['_id'] = data['identifier']
    return data


def _migrate_same_as_structure(data, client):
    """Changes sameAs string to schema:URL object."""
    same_as = data.get('same_as')

    if same_as and isinstance(same_as, str):
        data['same_as'] = {'@type': ['schema:URL'], 'url': same_as}

        if data.get('@context'):
            data['@context'].setdefault(
                'same_as', {
                    '@id': 'schema:sameAs',
                    '@type': 'schema:URL',
                    '@context': {
                        '@version': '1.1',
                        'url': 'schema:url',
                        'schema': 'http://schema.org/'
                    }
                }
            )

    return data


def _migrate_dataset_file_id(data, client):
    """Ensure dataset files have a fully qualified url as id."""
    host = 'localhost'
    if client:
        host = client.remote.get('host') or host
    host = os.environ.get('RENKU_DOMAIN') or host

    files = data.get('files', [])
    for file_ in files:
        if not file_['_id'].startswith('http'):
            file_['_id'] = 'https://{host}/{id}'.format(
                host=host, id=file_['_id']
            )

    return data


_PROJECT_FIELDS = {'_id', 'created', 'creator', 'name', 'updated', 'version'}

_DATASET_FIELDS = {
    '_id', '_label', '_project', 'based_on', 'created', 'creator',
    'date_published', 'description', 'files', 'identifier', 'in_language',
    'keywords', 'license', 'name', 'path', 'same_as', 'short_name', 'tags',
    'url', 'version'
}

_INITIAL_JSONLD_PROJECT_CONTEXT = {
    'schema': 'http://schema.org/',
    'prov': 'http://www.w3.org/ns/prov#',
    '@version': 1.1,
    'name': 'schema:name',
    'created': 'schema:dateCreated',
    'updated': 'schema:dateUpdated',
    'version': 'schema:schemaVersion',
    'creator': {
        '@id': 'schema:creator',
        '@context': {
            'schema': 'http://schema.org/',
            'prov': 'http://www.w3.org/ns/prov#',
            'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
            '@version': 1.1,
            'name': 'schema:name',
            'email': 'schema:email',
            'label': 'rdfs:label',
            'affiliation': 'schema:affiliation',
            'alternate_name': 'schema:alternateName',
            '_id': '@id'
        }
    },
    '_id': '@id'
}

_INITIAL_JSONLD_DATASET_CONTEXT = {
    'schema': 'http://schema.org/',
    '@version': 1.1,
    'prov': 'http://www.w3.org/ns/prov#',
    'wfprov': 'http://purl.org/wf4ever/wfprov#',
    'path': 'prov:atLocation',
    '_id': '@id',
    '_project': {
        '@id': 'schema:isPartOf',
        '@context': {
            'schema': 'http://schema.org/',
            'prov': 'http://www.w3.org/ns/prov#',
            '@version': 1.1,
            'name': 'schema:name',
            'created': 'schema:dateCreated',
            'updated': 'schema:dateUpdated',
            'version': 'schema:schemaVersion',
            'creator': {
                '@id': 'schema:creator',
                '@context': {
                    'schema': 'http://schema.org/',
                    'prov': 'http://www.w3.org/ns/prov#',
                    'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
                    '@version': 1.1,
                    'name': 'schema:name',
                    'email': 'schema:email',
                    'label': 'rdfs:label',
                    'affiliation': 'schema:affiliation',
                    'alternate_name': 'schema:alternateName',
                    '_id': '@id'
                }
            },
            '_id': '@id'
        }
    },
    'creator': {
        '@id': 'schema:creator',
        '@context': {
            'schema': 'http://schema.org/',
            'prov': 'http://www.w3.org/ns/prov#',
            'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
            '@version': 1.1,
            'name': 'schema:name',
            'email': 'schema:email',
            'label': 'rdfs:label',
            'affiliation': 'schema:affiliation',
            'alternate_name': 'schema:alternateName',
            '_id': '@id'
        }
    },
    'date_published': 'schema:datePublished',
    'description': 'schema:description',
    'identifier': 'schema:identifier',
    'in_language': {
        '@id': 'schema:inLanguage',
        '@context': {
            'schema': 'http://schema.org/',
            '@version': 1.1,
            'alternate_name': 'schema:alternateName',
            'name': 'schema:name'
        }
    },
    'keywords': 'schema:keywords',
    'based_on': 'schema:isBasedOn',
    'license': 'schema:license',
    'name': 'schema:name',
    'url': 'schema:url',
    'version': 'schema:version',
    'created': 'schema:dateCreated',
    'files': {
        '@id': 'schema:hasPart',
        '@context': {
            'schema': 'http://schema.org/',
            '@version': 1.1,
            'prov': 'http://www.w3.org/ns/prov#',
            'wfprov': 'http://purl.org/wf4ever/wfprov#',
            'path': 'prov:atLocation',
            '_id': '@id',
            '_project': {
                '@id': 'schema:isPartOf',
                '@context': {
                    'schema': 'http://schema.org/',
                    'prov': 'http://www.w3.org/ns/prov#',
                    '@version': 1.1,
                    'name': 'schema:name',
                    'created': 'schema:dateCreated',
                    'updated': 'schema:dateUpdated',
                    'version': 'schema:schemaVersion',
                    'creator': {
                        '@id': 'schema:creator',
                        '@context': {
                            'schema': 'http://schema.org/',
                            'prov': 'http://www.w3.org/ns/prov#',
                            'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
                            '@version': 1.1,
                            'name': 'schema:name',
                            'email': 'schema:email',
                            'label': 'rdfs:label',
                            'affiliation': 'schema:affiliation',
                            'alternate_name': 'schema:alternateName',
                            '_id': '@id'
                        }
                    },
                    '_id': '@id'
                }
            },
            'creator': {
                '@id': 'schema:creator',
                '@context': {
                    'schema': 'http://schema.org/',
                    'prov': 'http://www.w3.org/ns/prov#',
                    'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
                    '@version': 1.1,
                    'name': 'schema:name',
                    'email': 'schema:email',
                    'label': 'rdfs:label',
                    'affiliation': 'schema:affiliation',
                    'alternate_name': 'schema:alternateName',
                    '_id': '@id'
                }
            },
            'added': 'schema:dateCreated',
            'name': 'schema:name',
            'url': 'schema:url',
            'based_on': 'schema:isBasedOn'
        }
    },
    'tags': {
        '@id': 'schema:subjectOf',
        '@context': {
            'schema': 'http://schema.org/',
            '@version': 1.1,
            'name': 'schema:name',
            'description': 'schema:description',
            'commit': 'schema:location',
            'created': 'schema:startDate',
            'dataset': 'schema:about',
            '_id': '@id'
        }
    },
    'same_as': {
        '@id': 'schema:sameAs',
        '@context': {
            'schema': 'http://schema.org/',
            '@version': 1.1,
            'url': 'schema:url',
            '_id': '@id'
        }
    },
    'short_name': 'schema:alternateName'
}



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Show information about objects in current repository.

Siblings
~~~~~~~~

In situations when multiple outputs have been generated by a single
``renku run`` command, the siblings can be discovered by running
``renku show siblings PATH`` command.

Assume that the following graph represents relations in the repository.

.. code-block:: text

          D---E---G
         /     \
    A---B---C   F

Then the following outputs would be shown.

.. code-block:: console

   $ renku show siblings C
   C
   D
   $ renku show siblings G
   F
   G
   $ renku show siblings A
   A
   $ renku show siblings C G
   C
   D
   ---
   F
   G
   $ renku show siblings
   A
   ---
   B
   ---
   C
   D
   ---
   E
   ---
   F
   G

You can use the ``-f`` or ``--flat`` flag to output a flat list, as well as
the ``-v`` or ``--verbose`` flag to also output commit information.


Input and output files
~~~~~~~~~~~~~~~~~~~~~~

You can list input and output files generated in the repository by running
``renku show inputs`` and ``renku show outputs`` commands. Alternatively,
you can check if all paths specified as arguments are input or output files
respectively.

.. code-block:: console

   $ renku run wc < source.txt > result.wc
   $ renku show inputs
   source.txt
   $ renku show outputs
   result.wc
   $ renku show outputs source.txt
   $ echo $?  # last command finished with an error code
   1

"""

import click

from renku.core import errors
from renku.core.commands.client import pass_local_client
from renku.core.commands.graph import Graph
from renku.core.models.entities import Entity


@click.group()
def show():
    """Show information about objects in a current repository.

    NOTE: The command produces a machine-readable output.
    """
    pass


@show.command()
@click.option('--revision', default='HEAD')
@click.option('-f', '--flat', is_flag=True)
@click.option('-v', '--verbose', is_flag=True)
@click.argument(
    'paths', type=click.Path(exists=True, dir_okay=False), nargs=-1
)
@pass_local_client(requires_migration=True)
def siblings(client, revision, flat, verbose, paths):
    """Show siblings for given paths."""
    graph = Graph(client)
    nodes = graph.build(paths=paths, revision=revision)
    nodes = [n for n in nodes if not isinstance(n, Entity) or n.parent]

    sibling_sets = {frozenset([n]) for n in set(nodes)}
    for node in nodes:
        try:
            sibling_sets.add(frozenset(graph.siblings(node)))
        except (errors.InvalidOutputPath):
            # ignore nodes that aren't outputs if no path was supplied
            if paths:
                raise
            else:
                sibling_sets.discard({node})

    result_sets = []
    for candidate in sibling_sets:
        new_result = []

        for result in result_sets:
            if candidate & result:
                candidate |= result
            else:
                new_result.append(result)

        result_sets = new_result
        result_sets.append(candidate)

    result = [[sibling_name(graph, node, verbose) for node in r]
              for r in result_sets]

    if flat:
        click.echo('\n'.join({n for r in result for n in r}))
    else:
        click.echo('\n---\n'.join('\n'.join(r) for r in result))


@show.command()
@click.option('--revision', default='HEAD')
@click.argument(
    'paths',
    type=click.Path(exists=True, dir_okay=False),
    nargs=-1,
)
@pass_local_client(requires_migration=True)
@click.pass_context
def inputs(ctx, client, revision, paths):
    r"""Show inputs files in the repository.

    <PATHS>    Files to show. If no files are given all input files are shown.
    """
    from renku.core.models.provenance.activities import ProcessRun

    graph = Graph(client)
    paths = set(paths)
    nodes = graph.build(revision=revision)
    commits = {
        node.activity.commit if hasattr(node, 'activity') else node.commit
        for node in nodes
    }
    commits |= {
        node.activity.commit
        for node in nodes if hasattr(node, 'activity')
    }
    candidates = {(node.commit, node.path)
                  for node in nodes if not paths or node.path in paths}

    input_paths = set()

    for commit in commits:
        activity = graph.activities[commit]

        if isinstance(activity, ProcessRun):
            for usage in activity.qualified_usage:
                for entity in usage.entity.entities:
                    path = str((usage.client.path / entity.path).relative_to(
                        client.path
                    ))
                    usage_key = (entity.commit, entity.path)

                    if path not in input_paths and usage_key in candidates:
                        input_paths.add(path)

    click.echo('\n'.join(graph._format_path(path) for path in input_paths))
    ctx.exit(0 if not paths or len(input_paths) == len(paths) else 1)


@show.command()
@click.option('--revision', default='HEAD')
@click.argument(
    'paths',
    type=click.Path(exists=True, dir_okay=True),
    nargs=-1,
)
@pass_local_client(requires_migration=True)
@click.pass_context
def outputs(ctx, client, revision, paths):
    r"""Show output files in the repository.

    <PATHS>    Files to show. If no files are given all output files are shown.
    """
    graph = Graph(client)
    filter = graph.build(paths=paths, revision=revision)
    output_paths = graph.output_paths

    click.echo('\n'.join(graph._format_path(path) for path in output_paths))

    if paths:
        if not output_paths:
            ctx.exit(1)

        from renku.core.models.datastructures import DirectoryTree
        tree = DirectoryTree.from_list(item.path for item in filter)

        for output in output_paths:
            if tree.get(output) is None:
                ctx.exit(1)
                return


def sibling_name(graph, node, verbose=False):
    """Return the display name of a sibling."""
    name = graph._format_path(node.path)

    if verbose:
        name = '{} @ {}'.format(name, node.commit)

    return name


def _context_names():
    """Return list of valid context names."""
    from renku.core.models.jsonld import JSONLDMixin

    # NOTE: You should only pass a list or tuple of choices. Other iterables
    # (like generators) may lead to surprising results.
    return [cls.__name__ for cls in JSONLDMixin.__type_registry__.values()]


def print_context_names(ctx, param, value):
    """Print all possible types."""
    if not value or ctx.resilient_parsing:
        return
    click.echo('\n'.join(_context_names()))
    ctx.exit()


def _context_json(name):
    """Return JSON-LD string for given context name."""
    from renku.core.models.jsonld import JSONLDMixin

    for cls in JSONLDMixin.__type_registry__.values():
        if cls.__name__ == name:
            return {
                '@context': cls._jsonld_context,
                '@type': cls._jsonld_type,
            }


@show.command()
@click.argument(
    'names',
    type=click.Choice(_context_names()),
    nargs=-1,
)
@click.option(
    '--list',
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=print_context_names,
    help=print_context_names.__doc__,
)
def context(names):
    """Show JSON-LD context for repository objects."""
    import json

    contexts = [_context_json(name) for name in set(names)]
    if contexts:
        click.echo(
            json.dumps(
                contexts[0] if len(contexts) == 1 else contexts,
                indent=2,
            )
        )



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service templates view."""

from flask import Blueprint, request
from flask_apispec import marshal_with, use_kwargs
from marshmallow import EXCLUDE

from renku.core.commands.init import read_template_manifest
from renku.service.config import SERVICE_PREFIX
from renku.service.serializers.templates import ManifestTemplatesRequest, \
    ManifestTemplatesResponseRPC
from renku.service.views import result_response
from renku.service.views.cache import _project_clone
from renku.service.views.decorators import accepts_json, handle_base_except, \
    handle_git_except, handle_renku_except, handle_validation_except, \
    header_doc, requires_cache, requires_identity

TEMPLATES_BLUEPRINT_TAG = 'templates'
templates_blueprint = Blueprint(
    TEMPLATES_BLUEPRINT_TAG, __name__, url_prefix=SERVICE_PREFIX
)


@use_kwargs(ManifestTemplatesRequest, locations=['query'])
@marshal_with(ManifestTemplatesResponseRPC)
@header_doc('List templates in repositpry.', tags=(TEMPLATES_BLUEPRINT_TAG, ))
@templates_blueprint.route(
    '/templates.read_manifest',
    methods=['GET'],
    provide_automatic_options=False,
)
@handle_base_except
@handle_git_except
@handle_renku_except
@handle_validation_except
@requires_cache
@requires_identity
@accepts_json
def read_manifest_from_template(user, cache):
    """Read templates from the manifest file of a template repository."""
    project_data = ManifestTemplatesRequest().load({
        **user,
        **request.args,
    },
                                                   unknown=EXCLUDE)
    project = _project_clone(user, project_data)
    manifest = read_template_manifest(project.abs_path)

    return result_response(
        ManifestTemplatesResponseRPC(), {'templates': manifest}
    )



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""test KG against SHACL shape."""

import pyld

from renku.cli import cli
from renku.core.compat import Path
from renku.core.utils.shacl import validate_graph


def test_dataset_shacl(tmpdir, runner, project, client):
    """Test dataset metadata structure."""
    force_dataset_path = Path(
        __file__
    ).parent.parent.parent / 'fixtures' / 'force_dataset_shacl.json'

    force_datasetfile_path = Path(
        __file__
    ).parent.parent.parent / 'fixtures' / 'force_datasetfile_shacl.json'

    force_datasettag_path = Path(
        __file__
    ).parent.parent.parent / 'fixtures' / 'force_datasettag_shacl.json'

    runner.invoke(cli, ['dataset', 'create', 'dataset'])

    paths = []
    for i in range(3):
        new_file = tmpdir.join('file_{0}'.format(i))
        new_file.write(str(i))
        paths.append(str(new_file))

    # add data
    runner.invoke(
        cli,
        ['dataset', 'add', 'dataset'] + paths,
        catch_exceptions=False,
    )

    runner.invoke(
        cli,
        ['dataset', 'tag', 'dataset', '1.0'],
        catch_exceptions=False,
    )

    with client.with_dataset('dataset') as dataset:
        g = dataset.asjsonld()
        rdf = pyld.jsonld.to_rdf(
            g,
            options={
                'format': 'application/n-quads',
                'produceGeneralizedRdf': True
            }
        )

        r, _, t = validate_graph(rdf, shacl_path=str(force_dataset_path))
        assert r is True, t

        r, _, t = validate_graph(rdf, shacl_path=str(force_datasetfile_path))
        assert r is True, t

        r, _, t = validate_graph(rdf, shacl_path=str(force_datasettag_path))
        assert r is True, t

        r, _, t = validate_graph(rdf)
        assert r is True, t


def test_project_shacl(project, client):
    """Test project metadata structure."""
    from renku.core.models.provenance.agents import Person

    path = Path(
        __file__
    ).parent.parent.parent / 'fixtures' / 'force_project_shacl.json'

    project = client.project
    project.creator = Person(email='johndoe@example.com', name='Johnny Doe')

    g = project.asjsonld()
    rdf = pyld.jsonld.to_rdf(
        g,
        options={
            'format': 'application/n-quads',
            'produceGeneralizedRdf': False
        }
    )
    r, _, t = validate_graph(rdf, shacl_path=str(path))
    assert r is True, t

    r, _, t = validate_graph(rdf)
    assert r is True, t



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initial migrations."""
import os
import shutil
import urllib
import uuid
from pathlib import Path, posixpath
from urllib.parse import quote

from renku.core.models.datasets import Dataset
from renku.core.models.refs import LinkReference
from renku.core.utils.urls import url_to_string


def migrate(client):
    """Migration function."""
    _ensure_clean_lock(client)
    _do_not_track_lock_file(client)
    _migrate_datasets_pre_v0_3(client)
    _migrate_broken_dataset_paths(client)
    _fix_uncommitted_labels(client)
    _fix_dataset_files_urls(client)
    _fix_broken_dataset_file_project(client)
    _dataset_file_id_migration(client)
    _migrate_files_project(client)


def _ensure_clean_lock(client):
    """Make sure Renku lock file is not part of repository."""
    lock_file = client.path / '.renku.lock'
    try:
        lock_file.unlink()
    except FileNotFoundError:
        pass


def _do_not_track_lock_file(client):
    """Add lock file to .gitingore if not already exists."""
    # Add lock file to .gitignore.
    lock_file = '.renku.lock'
    gitignore = client.path / '.gitignore'
    if lock_file not in gitignore.read_text():
        gitignore.open('a').write('\n{0}\n'.format(lock_file))


def _migrate_datasets_pre_v0_3(client):
    """Migrate datasets from Renku 0.3.x."""

    def _dataset_pre_0_3(client):
        """Return paths of dataset metadata for pre 0.3.4."""
        project_is_pre_0_3 = int(client.project.version) < 2
        if project_is_pre_0_3:
            return (client.path / 'data').rglob(client.METADATA)
        return []

    for old_path in _dataset_pre_0_3(client):
        name = str(old_path.parent.relative_to(client.path / 'data'))

        dataset = Dataset.from_yaml(old_path, client=client)
        new_path = (client.renku_datasets_path / dataset.uid / client.METADATA)
        new_path.parent.mkdir(parents=True, exist_ok=True)

        with client.with_metadata(read_only=True) as meta:
            for module in client.repo.submodules:
                if Path(module.url).name == meta.name:
                    module.remove()

        for file_ in dataset.files:
            if not Path(file_.path).exists():
                expected_path = (
                    client.path / 'data' / dataset.name / file_.path
                )
                if expected_path.exists():
                    file_.path = expected_path.relative_to(client.path)

        dataset.__reference__ = new_path.relative_to(client.path)
        dataset.to_yaml()

        Path(old_path).unlink()
        ref = LinkReference.create(
            client=client,
            name='datasets/{0}'.format(name),
            force=True,
        )
        ref.set_reference(new_path)


def _migrate_broken_dataset_paths(client):
    """Ensure all paths are using correct directory structure."""
    for dataset in client.datasets.values():
        dataset_path = client.path / dataset.path

        expected_path = (
            client.renku_datasets_path /
            Path(quote(dataset.identifier, safe=''))
        )

        # migrate the refs
        ref = LinkReference.create(
            client=client,
            name='datasets/{0}'.format(dataset.short_name),
            force=True,
        )
        ref.set_reference(expected_path / client.METADATA)

        if not dataset_path.exists():
            dataset_path = (
                client.renku_datasets_path / uuid.UUID(dataset.identifier).hex
            )

        if not expected_path.exists():
            shutil.move(dataset_path, expected_path)
            dataset.path = expected_path
            dataset.__reference__ = expected_path / client.METADATA

        for file_ in dataset.files:
            file_path = Path(file_.path)
            if not file_path.exists() and file_.path.startswith('..'):
                new_path = (
                    client.renku_datasets_path / dataset.uid / file_path
                ).resolve().relative_to(client.path)

                file_.path = new_path

                _, commit, _ = client.resolve_in_submodules(
                    client.find_previous_commit(file_.path, revision='HEAD'),
                    file_.path,
                )
                host = client.remote.get('host') or 'localhost'
                host = os.environ.get('RENKU_DOMAIN') or host

                # always set the id by the identifier
                file_._id = urllib.parse.urljoin(
                    'https://{host}'.format(host=host),
                    posixpath.join(
                        '/blob/{hexsha}/{path}'.format(
                            hexsha=commit.hexsha, path=new_path
                        )
                    )
                )
                file_._label = '{}@{}'.format(new_path, commit.hexsha)

        dataset.to_yaml()


def _fix_uncommitted_labels(client):
    """Ensure files have correct label instantiation."""
    for dataset in client.datasets.values():
        for file_ in dataset.files:
            try:
                _, commit, _ = client.resolve_in_submodules(
                    client.find_previous_commit(file_.path, revision='HEAD'),
                    file_.path,
                )
                file_.commit = commit
                if 'UNCOMMITTED' in file_._label or '@' not in file_._label:
                    file_._label = file_.default_label()
                    file_._id = file_.default_id()
            except KeyError:
                pass
        dataset.to_yaml()


def _fix_dataset_files_urls(client):
    """Ensure dataset files have correct url format."""
    for dataset in client.datasets.values():
        for file_ in dataset.files:
            if file_.url:
                file_.url = url_to_string(file_.url)

        dataset.to_yaml()


def _fix_broken_dataset_file_project(client):
    """Ensure project is correctly set on ``DatasetFile``."""
    for dataset in client.datasets.values():
        for file_ in dataset.files:
            if not file_._project or 'NULL/NULL' in file_._project._id:
                file_._project = client.project

        dataset.to_yaml()


def _dataset_file_id_migration(client):
    """Ensure dataset files have a fully qualified url."""
    for dataset in client.datasets.values():
        for file_ in dataset.files:
            if not file_._id.startswith('https'):
                file_._id = file_.default_id()

        dataset.to_yaml()


def _migrate_files_project(client):
    """Ensure dataset files have correct project."""
    for dataset in client.datasets.values():
        for file_ in dataset.files:
            file_._project = dataset._project

        dataset.to_yaml()
